#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


# cython: language_level=3
# coding=utf-8





from cpython cimport bool
from libc.stdint cimport int64_t

cdef extern from "Python.h":
    object PyLong_FromLong(long v)


cdef class IDGeneratorInt:
    cdef public int64_t counter

    def __init__(self):
        self.counter = 0

    cpdef int64_t get_new_id(self):
        cdef int64_t current_counter = self.counter
        self.counter += 1
        return current_counter

    cpdef remove_id(self, int64_t id_to_be_removed):
        pass

    cpdef clear(self):
        self.counter = 0

    def __call__(self):
        return self.get_new_id()


cdef class IDGeneratorUUID:
    cdef public int64_t counter
    cdef public object uuid

    def __init__(self):
        self.counter = 0
        self.uuid = __import__('uuid')  # Importing uuid in a way compatible with Cython

    cpdef get_new_id(self):
        cdef int64_t seq = self.counter
        self.counter += 1
        # Directly use the Python object without C type declaration
        return self.uuid.uuid1(clock_seq=seq).hex

    cpdef remove_id(self, id_to_be_removed):
        pass

    cpdef clear(self):
        self.counter = 0

    def __call__(self):
        return self.get_new_id()


cdef class IDGeneratorFakeUUID:
    cdef public int64_t counter

    def __init__(self):
        self.counter = 0

    cpdef get_new_id(self):
        cdef int64_t current_counter = self.counter
        self.counter += 1
        return str(PyLong_FromLong(current_counter))

    cpdef remove_id(self, object id_to_be_removed):
        pass

    cpdef clear(self):
        self.counter = 0

    def __call__(self):
        return self.get_new_id()


# cdef class IDGenerator:
#     cdef public object gen
#     cdef public object generator_type

#     def __init__(self, object generator_type = None):
#         self.gen = None
#         if generator_type is None:
#             self.generator_type = 0
#         else:
#             self.generator_type = generator_type.value

#         # Incorporate the UUID presence check here
#         uuid_present = self._check_uuid_availability()

#         if 0 == self.generator_type:
#             self.gen = IDGeneratorInt()
#         elif 1 == self.generator_type:
#             if uuid_present:
#                 self.gen = IDGeneratorUUID()
#             else:
#                 self.gen = IDGeneratorFakeUUID()
#         elif 2 == self.generator_type:
#             raise NotImplementedError
#         else:
#             raise RuntimeError('Wrong `generator_type` value')

#     cdef bool _check_uuid_availability(self):
#         try:
#             __import__('uuid')
#             return True
#         except ImportError:
#             return False

#     cpdef get_new_id(self):
#         return self.gen.get_new_id()

#     cpdef remove_id(self, id_to_be_removed):
#         self.gen.remove_id(id_to_be_removed)

#     cpdef clear(self):
#         self.gen.clear()

#     def __call__(self):
#         return self.gen.get_new_id()


cdef bool _check_uuid_availability():
    try:
        __import__('uuid')
        return True
    except ImportError:
        return False


cpdef IDGenerator(object generator_type = None):
    gen = None
    if generator_type is None:
        generator_type = 0
    else:
        generator_type = generator_type.value

    # Incorporate the UUID presence check here
    uuid_present = _check_uuid_availability()

    if 0 == generator_type:
        return IDGeneratorInt()
    elif 1 == generator_type:
        if uuid_present:
            return IDGeneratorUUID()
        else:
            return IDGeneratorFakeUUID()
    elif 2 == generator_type:
        raise NotImplementedError
    else:
        raise RuntimeError('Wrong `generator_type` value')



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['SHGetKnownFolderPathError', 'SHGetKnownFolderPath']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


import os
from ctypes import POINTER, byref, c_wchar_p
from cengal.ctypes_tools.types import GUID
from cengal.ctypes_tools.constants import KnownfolderidConstants, S_OK
from cengal.ctypes_tools.function_prototypes import SHGetKnownFolderPath as proto_SHGetKnownFolderPath, CoTaskMemFree
from .result_api_exceptions import CtypesToolsResultAPIError
from cengal.file_system.directory_manager import ensure_dir as ensure_dir_exists
from typing import Union, Sequence, Any


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class CtypesToolsWinResultAPIError(CtypesToolsResultAPIError):
    pass


class SHGetKnownFolderPathError(CtypesToolsWinResultAPIError):
    pass


def SHGetKnownFolderPath(dir_guid: GUID):
    """Example:

        from cengal.ctypes_tools.constants import KnownfolderidConstants
        dir_path: str = SHGetKnownFolderPath(KnownfolderidConstants.FOLDERID_LocalAppData)

    Args:
        dir_guid (GUID): One of KnownfolderidConstants constants. `from cengal.ctypes_tools.constants import KnownfolderidConstants`

    Raises:
        SHGetKnownFolderPathError: _description_

    Returns:
        _type_: _description_
    """
    result_path = None
    path_ptr = c_wchar_p()
    result = proto_SHGetKnownFolderPath(byref(dir_guid), 0, None, byref(path_ptr))
    if result != S_OK:
        raise SHGetKnownFolderPathError(result)
        
    try:
        result_path = path_ptr.value
    finally:
        if path_ptr:
            CoTaskMemFree(path_ptr)
    
    return result_path



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.math.geometry.ellipse import Ellipse

"""Results for Ellipse(5, 1):
    21.205750411731103
    21.010026651559343
    23.876104167282428
    21.188826246073443
    21.010185796052752
    21.010070798076192
    21.0100598492146
    21.01005354033284
    21.010046444646637
    21.010045234814886
    (21.010044539689165, 533)
    (21.010044539689044, 4096)
"""

"""Results for Ellipse(500, 1):
    1887.3117866440682
    1999.3138980764822
    2356.1976317849985
    2048.179440784419
    2004.993211462262
    2003.847518534776
    2003.5744652543665
    2003.3377375221762
    2002.785086377765
    2002.5086163219212
    (2000.8769360634774, 581)
    (2000.1414074088598, 4096)
"""

el = Ellipse(5, 1)
print(el.perimeter__best_of__euler__and__matt_parker__lazy)
print(el.perimeter__ramanujan_2)
print(el.perimeter__infinite_sum__iter_lim(1))
print(el.perimeter__infinite_sum__iter_lim(10))
print(el.perimeter__infinite_sum__iter_lim(100))
print(el.perimeter__infinite_sum__iter_lim(130))
print(el.perimeter__infinite_sum__iter_lim(140))  # For Ellipse(5, 1): Diviation is similar to perimeter__ramanujan_2 but with opposite sign (perimeter__ramanujan_2 gives number smaller than needed and this gives number greater than needed)
print(el.perimeter__infinite_sum__iter_lim(150))
print(el.perimeter__infinite_sum__iter_lim(180))
print(el.perimeter__infinite_sum__iter_lim(200))
print(el.perimeter__infinite_sum__time_lim(0.100, True))
print(el.perimeter__infinite_sum__time_lim(10.0, True))




#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cengal.code_inspection.line_tracer import LineTracer
from cengal.data_generation.id_generator import IDGenerator
from cengal.introspection.inspect import intro_func_repr, intro_func_repr_limited, get_multistr_of_data_value
from enum import Enum, IntEnum
from typing import Callable, Optional, Any, Set
try:
    import rich
    RICH_PRESENT = True
except ImportError:
    RICH_PRESENT = False

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class CodeStartReplType(IntEnum):
    general = 0
    general_verbose = 1
    limited = 2
    limited_verbose = 3


class LineType(IntEnum):
    previous_line = 0
    current_line = 1
    next_line = 2
    exact_line = 3
    relative_line = 4


class OutputFields(IntEnum):
    trace_name = 0
    file_name = 1
    line = 2
    func_name = 3
    code_line = 4
    result = 5
    new_line_after_end = 6
    start_short_name = 7


class AutoLineTracer:
    def __init__(self, code_start_repl_type: CodeStartReplType = CodeStartReplType.general, print_allowed: bool = True, rich_allowed: bool = True, *args, **kwargs):
        self.code_start_repl_type: CodeStartReplType = code_start_repl_type
        self.print_allowed: bool = print_allowed
        self.rich_allowed: bool = rich_allowed
        if CodeStartReplType.general == code_start_repl_type:
            self.code_start_repl = self._start_impl_general
        elif CodeStartReplType.general_verbose == code_start_repl_type:
            self.code_start_repl = self._start_impl_general_verbose
        elif CodeStartReplType.limited == code_start_repl_type:
            self.code_start_repl = self._start_impl_limited
        elif CodeStartReplType.limited_verbose == code_start_repl_type:
            self.code_start_repl = self._start_impl_limited_verbose
        
        self.lt: LineTracer = LineTracer(*args, **kwargs)
        self._index: IDGenerator = IDGenerator()
        self.line_template: str = '#{index:<4n}| <[{name}]> | <file \'{file_name}\' line {line}>.{func_name}()\n\t| {code_line}'
        self.line_template_name_less: str = '#{index:<4n}| <file \'{file_name}\' line {line}>.{func_name}()\n\t| {code_line}'
        # self.start_template: str = '#{index:<4n}| <+[{name}]+>'
        self.start_template: str = '#{index:<4n}| <+[{short_name}]+>\n\t|<+[{name}]+>'
        self.end_template: str = '#{index:<4n}| <-[{short_name}]->\n\t|<-[{name}]->'

        self.s = self.start
        self.e = self.end
        self.p = self.previous_line
        self.c = self.current_line
        self.n = self.next_line

        self.ps = self.print_start
        self.pe = self.print_end
        self.pp = self.print_previous_line
        self.pc = self.print_current_line
        self.pcptv = self.print_current_line_pp_type_value
        self.pcpt = self.print_current_line_pp_value
        self.pn = self.print_next_line
    
    @property
    def index(self):
        return self._index()

    def _start_impl_general(self, depth: int = 1):
        return intro_func_repr(False, depth + 1)

    def _start_impl_general_verbose(self, depth: int = 1):
        return intro_func_repr(True, depth + 1)

    def _start_impl_limited(self, depth: int = 1):
        return intro_func_repr_limited(False, depth + 1)

    def _start_impl_limited_verbose(self, depth: int = 1):
        return intro_func_repr_limited(True, depth + 1)
    
    def start(self, depth: int = 1):
        short_name = self._start_impl_limited(depth + 1)
        name = self.code_start_repl(depth + 1)
        return self.start_template.format(index=self.index, short_name=short_name, name=name)

    def print_start(self, depth: int = 1):
        if self.print_allowed:
            print(self.start(depth + 1) + '\n')

    def end(self, depth: int = 1):
        short_name = self._start_impl_limited(depth + 1)
        name = self.code_start_repl(depth + 1)
        return self.end_template.format(index=self.index, short_name=short_name, name=name)

    def print_end(self, depth: int = 1):
        if self.print_allowed:
            print(self.end(depth + 1) + '\n')

    def previous_line(self, name: Optional[str] = None, depth: int = 1):
        filename, function_name, line_number, lines, index = self.lt.trace(depth + 1)
        lines = lines.strip()
        if name is None:
            return self.line_template_name_less.format(index=self.index, file_name=filename, line=line_number, func_name=function_name, code_line=lines)
        else:
            return self.line_template.format(index=self.index, name=name, file_name=filename, line=line_number, func_name=function_name, code_line=lines)

    def print_previous_line(self, name: Optional[str] = None, depth: int = 1):
        if self.print_allowed:
            print(self.previous_line(name, depth + 1) + '\n')

    def current_line(self, line_result, name: Optional[str] = None, line_type: LineType = LineType.current_line, line_num: Optional[int] = None, 
                     output_fields: Optional[Set[OutputFields]] = None, 
                     depth: int = 1):
        if output_fields is None:
            output_fields = {
                OutputFields.trace_name, 
                OutputFields.file_name,
                OutputFields.line,
                OutputFields.func_name,
                OutputFields.code_line,
                OutputFields.result,
            }

        if LineType.current_line == line_type:
            filename, function_name, line_number, lines, index = self.lt.trace_self(depth + 1)
        elif LineType.exact_line == line_type:
            filename, function_name, line_number, lines, index = self.lt.trace_exact(line_num, depth + 1)
        elif LineType.relative_line == line_type:
            filename, function_name, line_number, lines, index = self.lt.trace_relative(line_num, depth + 1)
        elif LineType.previous_line == line_type:
            filename, function_name, line_number, lines, index = self.lt.trace(depth + 1)
        elif LineType.next_line == line_type:
            filename, function_name, line_number, lines, index = self.lt.trace_next(depth + 1)

        lines = lines.strip()
        if name is None:
            result = self.line_template_name_less.format(index=self.index, file_name=filename, line=line_number, func_name=function_name, code_line=lines)
        else:
            result = self.line_template.format(index=self.index, name=name, file_name=filename, line=line_number, func_name=function_name, code_line=lines)
        
        if OutputFields.result in output_fields:
            result += f'\n\t\t| {line_result}'
        
        return result

    def print_current_line(self, line_result, name: Optional[str] = None, line_type: LineType = LineType.current_line, line_num: Optional[int] = None, output_fields: Optional[Set[OutputFields]] = None, depth: int = 1) -> Any:
        if output_fields is None:
            output_fields = {
                OutputFields.trace_name, 
                OutputFields.file_name,
                OutputFields.line,
                OutputFields.func_name,
                OutputFields.code_line,
                OutputFields.result,
                OutputFields.new_line_after_end,
            }

        if self.print_allowed:
            if RICH_PRESENT and self.rich_allowed:
                from rich.console import Console
                from rich.syntax import Syntax

                if LineType.current_line == line_type:
                    filename, function_name, line_number, lines, index = self.lt.trace_self(depth + 1)
                elif LineType.exact_line == line_type:
                    filename, function_name, line_number, lines, index = self.lt.trace_exact(line_num, depth + 1)
                elif LineType.relative_line == line_type:
                    filename, function_name, line_number, lines, index = self.lt.trace_relative(line_num, depth + 1)
                elif LineType.previous_line == line_type:
                    filename, function_name, line_number, lines, index = self.lt.trace(depth + 1)
                elif LineType.next_line == line_type:
                    filename, function_name, line_number, lines, index = self.lt.trace_next(depth + 1)

                # lines = lines.strip()
                syntax = Syntax(lines, "python", theme="monokai", line_numbers=True, start_line=line_number)
                console = Console()
                cl = self.current_line(line_result, name, line_type, line_num, output_fields, depth + 1)
                clines = cl.split('\n')
                # print(clines[0])
                console.print(clines[0])
                console.print(syntax)
                clines_rest = clines[2:]
                for cline in clines_rest:
                    console.print(cline)
                
                if OutputFields.new_line_after_end in output_fields:
                    print()
                    # console.print(clines[3])
                    # print(clines[2])
            else:
                if OutputFields.new_line_after_end in output_fields:
                    ending = '\n'
                else:
                    ending = ''

                print(self.current_line(line_result, name, line_type, line_num, output_fields, depth + 1) + ending)
        
        return line_result

    def print_current_line_pp_type_value(self, line_result, name: Optional[str] = None, line_type: LineType = LineType.current_line, line_num: Optional[int] = None, output_fields: Optional[Set[OutputFields]] = None, depth: int = 1) -> Any:
        multistr_data_value: str = get_multistr_of_data_value(line_result, 3).lstrip(' \t')
        str_template = f'{type(line_result)}\n\t\t| {multistr_data_value}'
        self.print_current_line(str_template, name, line_type, line_num, output_fields, depth + 1)
        return line_result

    def print_current_line_pp_value(self, line_result, name: Optional[str] = None, line_type: LineType = LineType.current_line, line_num: Optional[int] = None, output_fields: Optional[Set[OutputFields]] = None, depth: int = 1) -> Any:
        str_template = get_multistr_of_data_value(line_result, 3).lstrip(' \t')
        self.print_current_line(str_template, name, line_type, line_num, output_fields, depth + 1)
        return line_result
    
    def next_line(self, name: Optional[str] = None, depth: int = 1):
        filename, function_name, line_number, lines, index = self.lt.trace_next(depth + 1)
        lines = lines.strip()
        if name is None:
            return self.line_template_name_less.format(index=self.index, file_name=filename, line=line_number, func_name=function_name, code_line=lines)
        else:
            return self.line_template.format(index=self.index, name=name, file_name=filename, line=line_number, func_name=function_name, code_line=lines)

    def print_next_line(self, name: Optional[str] = None, depth: int = 1):
        if self.print_allowed:
            print(self.next_line(name, depth + 1) + '\n')
    
    def __call__(self, depth: int = 1):
        _, _, prev_line_number, prev_lines, _ = self.lt.trace(depth + 1)
        filename, function_name, next_line_number, next_lines, _ = self.lt.trace_next(depth + 1)
        return filename, function_name, prev_line_number, next_line_number, prev_lines, next_lines


auto_line_tracer__general = AutoLineTracer(CodeStartReplType.general)
auto_line_tracer__general_verbose = AutoLineTracer(CodeStartReplType.general_verbose)
auto_line_tracer__limited = AutoLineTracer(CodeStartReplType.limited)
auto_line_tracer__limited_verbose = AutoLineTracer(CodeStartReplType.limited_verbose)

altg = auto_line_tracer__general
altgv = auto_line_tracer__general_verbose
altl = auto_line_tracer__limited
altlv = auto_line_tracer__limited_verbose

alt = altg


def trace_self__general(line_result, name: Optional[str] = None, line_type: LineType = LineType.current_line, line_num: Optional[int] = None, output_fields: Optional[Set[OutputFields]] = None, depth: int = 1) -> Any:
    return auto_line_tracer__general.print_current_line_pp_type_value(line_result, name, line_type, line_num, output_fields, depth + 1)


def trace_self__general_verbose(line_result, name: Optional[str] = None, line_type: LineType = LineType.current_line, line_num: Optional[int] = None, output_fields: Optional[Set[OutputFields]] = None, depth: int = 1) -> Any:
    return auto_line_tracer__general_verbose.print_current_line_pp_type_value(line_result, name, line_type, line_num, output_fields, depth + 1)


def trace_self__limited(line_result, name: Optional[str] = None, line_type: LineType = LineType.current_line, line_num: Optional[int] = None, output_fields: Optional[Set[OutputFields]] = None, depth: int = 1) -> Any:
    return auto_line_tracer__limited.print_current_line_pp_type_value(line_result, name, line_type, line_num, output_fields, depth + 1)


def trace_self__limited_verbose(line_result, name: Optional[str] = None, line_type: LineType = LineType.current_line, line_num: Optional[int] = None, output_fields: Optional[Set[OutputFields]] = None, depth: int = 1) -> Any:
    return auto_line_tracer__limited_verbose.print_current_line_pp_type_value(line_result, name, line_type, line_num, output_fields, depth + 1)


tsg = trace_self__general
tsgv = trace_self__general_verbose
tsl = trace_self__limited
tslv = trace_self__limited_verbose

ts = tsg


def fake_trace_self(line_result, name: Optional[str] = None, line_type: LineType = LineType.current_line, line_num: Optional[int] = None, output_fields: Optional[Set[OutputFields]] = None, depth: int = 1):
    return line_result


fts = fake_trace_self



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['await_coro_fast', 'await_coro', 'await_coro_prim', 'asyncio_coro', 'await_task_fast', 'await_task', 'await_task_prim', 
           'cs_awaitable', 
           'RunSchedulerInAsyncioLoop',
           'coro_interfaces_arg_manager', 'create_task', 'create_task_in_thread_pool', 'task_in_thread_pool']

import sys
import asyncio
from inspect import signature, Signature, Parameter, isawaitable
from cengal.parallel_execution.coroutines.coro_scheduler import *
# from cengal.parallel_execution.coroutines.coro_standard_services import *
from cengal.code_flow_control.args_manager import ArgsManager, EArgs
from enum import Enum
from functools import partial
from typing import Union, Optional, Callable, Awaitable, Any, Coroutine
from cengal.time_management.sleep_tools import get_min_sleep_interval, try_sleep, get_usable_min_sleep_interval, get_countable_delta_time
from cengal.introspection.inspect import get_exception, get_exception_tripple, is_async
from cengal.parallel_execution.coroutines.coro_standard_services_internal_lib.service_with_a_direct_request import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro

from functools import wraps, update_wrapper


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def _async_coro_wrapper(interface: Interface, future: asyncio.Future, coro_worker: Worker, *args, **kwargs):
    coro_worker_result = None
    try:
        if __debug__: dlog(f'λ wrapper => {func_info(coro_worker)}')
        coro_worker_result = coro_worker(interface, *args, **kwargs)
    except:
        if __debug__: dlog(f'λ wrapper (Exception) <= {repr(coro_worker)}')
        if not future.cancelled():
            ex_type, ex_value, ex_traceback = get_exception_tripple()
            if __debug__: dlog(ex_type, ex_value, ex_traceback)
            ex_value = ex_value.with_traceback(ex_traceback)
            future.set_exception(ex_value)
    else:
        if __debug__: dlog(f'λ wrapper <= {repr(coro_worker)}')
        if not future.cancelled():
            future.set_result(coro_worker_result)


async def _awaitable_async_coro_wrapper(interface: Interface, future: asyncio.Future, coro_worker: Worker, *args, **kwargs):
    coro_worker_result = None
    try:
        if __debug__: dlog(f'aλ wrapper => {func_info(coro_worker)}')
        
        # # TODO: fix required!
        # coro_worker_result = await interface(RunCoro, CoroType.awaitable, coro_worker, *args, **kwargs)
        # TODO: possible fix:
        coro_worker_result = await coro_worker(interface, *args, **kwargs)
    except:
        if __debug__: dlog(f'aλ wrapper (Exception) <= {repr(coro_worker)}')
        if not future.cancelled():
            ex_type, ex_value, ex_traceback = get_exception_tripple()
            if __debug__: dlog(ex_type, ex_value, ex_traceback)
            ex_value = ex_value.with_traceback(ex_traceback)
            future.set_exception(ex_value)
    else:
        if __debug__: dlog(f'aλ wrapper <= {repr(coro_worker)}')
        if not future.cancelled():
            future.set_result(coro_worker_result)


def await_coro_fast(loop: asyncio.AbstractEventLoop,
                    scheduler: CoroSchedulerType, coro_type: Optional[CoroType], coro_worker: Worker, *args, **kwargs
                    ) -> asyncio.Future:
    coro_type = coro_type or CoroType.auto
    if CoroType.auto == coro_type:
        coro_type = find_coro_type(coro_worker)
    
    future = loop.create_future()
    if CoroType.awaitable == coro_type:
        put_request_to_service_with_context(get_interface_and_loop_with_explicit_loop(scheduler), PutCoro, ExplicitWorker(coro_type, _awaitable_async_coro_wrapper), future, coro_worker, *args, **kwargs)
    elif CoroType.greenlet == coro_type:
        put_request_to_service_with_context(get_interface_and_loop_with_explicit_loop(scheduler), PutCoro, ExplicitWorker(coro_type, _async_coro_wrapper), future, coro_worker, *args, **kwargs)
    else:
        raise NotImplementedError
    
    return future


def await_coro(scheduler: CoroSchedulerType, coro_worker: Worker, *args, **kwargs) -> asyncio.Future:
    return await_coro_fast(asyncio.get_event_loop(), scheduler, None, coro_worker, *args, **kwargs)


def await_coro_prim(coro_worker: Worker, *args, **kwargs) -> asyncio.Future:
    """Tries to use primary coro scheduler (must be set before usage)

    Args:
        coro_worker (Worker): _description_

    Returns:
        asyncio.Future: _description_
    """
    return await_coro_fast(asyncio.get_event_loop(), available_coro_scheduler(), CoroType.auto, coro_worker, *args, **kwargs)


def asyncio_coro(coro_worker: Worker) -> Coroutine:
    """Decorator. Without arguments. Gives an ability to await any decorated Cengal coroutine from the async code

    Args:
        coro_worker (Worker): _description_

    Returns:
        Coroutine: _description_
    """    
    async def wrapper(*args, **kwargs):
        return await await_coro_prim(coro_worker, *args, **kwargs)
    
    coro_worker_sign: Signature = signature(coro_worker)
    wrapper.__signature__ = coro_worker_sign.replace(parameters=tuple(coro_worker_sign.parameters.values())[1:], return_annotation=coro_worker_sign.return_annotation)
    return wrapper


def _async_task_runner_coro_worker(interface: Interface, service_type: ServiceType, *args, **kwargs):
    if __debug__: dlog(f'λ wrapper <_async_task_runner_coro_worker> => interface({service_type}, {args}, {kwargs})')
    result = interface(service_type, *args, **kwargs)
    if __debug__: dlog(f'λ wrapper <_async_task_runner_coro_worker> <= interface({service_type}, {args}, {kwargs}): {result}')
    return result


def await_task_fast(loop: asyncio.AbstractEventLoop,
                    scheduler: CoroSchedulerType, service_type: ServiceType, *args, **kwargs
                    ) -> asyncio.Future:
    future = loop.create_future()
    put_request_to_service_with_context(get_interface_and_loop_with_explicit_loop(scheduler), PutCoro, ExplicitWorker(CoroType.greenlet, _async_coro_wrapper), future, _async_task_runner_coro_worker, service_type, *args, **kwargs)
    return future


def await_task(scheduler: CoroSchedulerType, service_type: ServiceType, *args, **kwargs) -> asyncio.Future:
    return await_task_fast(asyncio.get_event_loop(), scheduler, service_type, *args, **kwargs)


def await_task_prim(service_type: ServiceType, *args, **kwargs) -> asyncio.Future:
    """Tries to use primary coro scheduler (must be set before usage)

    Args:
        service_type (ServiceType): _description_

    Returns:
        asyncio.Future: _description_
    """
    return await_task_fast(asyncio.get_event_loop(), available_coro_scheduler(), service_type, *args, **kwargs)


def cs_awaitable(coro_worker: Callable) -> Callable:
    """Decorator. Without arguments. Makes any Cengal coro awaitable from the async code (like coroutines in asyncio)
    Example:
        from cengal.parallel_execution.coroutines.coro_scheduler import cs_acoro
        from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import RunCoro
        from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro
        from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
        import asyncio

        @cs_awaitable
        def my_coro_g(a: str, b: str) -> str:
            i: Interface = current_interface()
            i(Sleep, 0.1)

            return a + b

        @cs_awaitable
        async def my_coro_a(a: str, b: str) -> str:
            i: Interface = current_interface()
            await i(Sleep, 0.05)

            await asyncio.sleep(0.05)

            return a + b

        @cs_acoro
        async def my_coro_a_implicit(a: str, b: str) -> str:
            i: Interface = current_interface()
            await i(Sleep, 0.05)

            await asyncio.sleep(0.05)

            return a + b

        async def my_coro_a_explicit(i: Interface, a: str, b: str) -> str:
            await i(Sleep, 0.05)

            await asyncio.sleep(0.05)

            return a + b
        
        async def my_coro_asyncio(a: str, b: str) -> str:
            await asyncio.sleep(0.1)

            return a + b
        
        @cs_awaitable
        async def amain():
            i: Interface = current_interface()

            # await

            print(await my_coro_g('a', 'b'))
            print(await my_coro_a('a', 'b'))
            print(await my_coro_a_implicit('a', 'b'))
            print(await my_coro_a_explicit(i, 'a', 'b'))
            print(await my_coro_asyncio('a', 'b'))

            # RunCoro

            print(await i(RunCoro, my_coro_g('a', 'b')))
            print(await i(RunCoro, my_coro_g, 'a', 'b'))

            print(await i(RunCoro, my_coro_a('a', 'b')))
            print(await i(RunCoro, my_coro_a, 'a', 'b'))
            
            print(await i(RunCoro, my_coro_a_implicit('a', 'b')))
            print(await i(RunCoro, my_coro_a_implicit, 'a', 'b'))
            
            print(await i(RunCoro, my_coro_a_explicit, 'a', 'b'))
            
            print(await i(RunCoro, my_coro_asyncio('a', 'b')))
            print(await i(RunCoro, cs_acoro(my_coro_asyncio), 'a', 'b'))

            # PutCoro

            await i(PutCoro, my_coro_g('a', 'b'))
            await i(PutCoro, my_coro_g, 'a', 'b')

            await i(PutCoro, my_coro_a('a', 'b'))
            await i(PutCoro, my_coro_a, 'a', 'b')
            
            await i(PutCoro, my_coro_a_implicit('a', 'b'))
            await i(PutCoro, my_coro_a_implicit, 'a', 'b')
            
            await i(PutCoro, my_coro_a_explicit, 'a', 'b')
            
            await i(PutCoro, my_coro_asyncio('a', 'b'))
            await i(PutCoro, cs_acoro(my_coro_asyncio), 'a', 'b')
        
        @cs_awaitable
        def main():
            i: Interface = current_interface()

            # RunCoro

            print(i(RunCoro, my_coro_g('a', 'b')))
            print(i(RunCoro, my_coro_g, 'a', 'b'))

            print(i(RunCoro, my_coro_a('a', 'b')))
            print(i(RunCoro, my_coro_a, 'a', 'b'))
            
            print(i(RunCoro, my_coro_a_implicit('a', 'b')))
            print(i(RunCoro, my_coro_a_implicit, 'a', 'b'))
            
            print(i(RunCoro, my_coro_a_explicit, 'a', 'b'))
            
            print(i(RunCoro, my_coro_asyncio('a', 'b')))
            print(i(RunCoro, cs_acoro(my_coro_asyncio), 'a', 'b'))

            # PutCoro

            i(PutCoro, my_coro_g('a', 'b'))
            i(PutCoro, my_coro_g, 'a', 'b')

            i(PutCoro, my_coro_a('a', 'b'))
            i(PutCoro, my_coro_a, 'a', 'b')
            
            i(PutCoro, my_coro_a_implicit('a', 'b'))
            i(PutCoro, my_coro_a_implicit, 'a', 'b')
            
            i(PutCoro, my_coro_a_explicit, 'a', 'b')
            
            i(PutCoro, my_coro_asyncio('a', 'b'))
            i(PutCoro, cs_acoro(my_coro_asyncio), 'a', 'b')

    Args:
        coro_worker (Worker): _description_

    Returns:
        Coroutine: _description_
    """
    if is_async(coro_worker):
        @wraps(coro_worker)
        async def wrapper(*args, **kwargs) -> Any:
            if isawaitable(coro_worker):
                return await coro_worker
            else:
                return await coro_worker(*args, **kwargs)
        
        wrapper: coro_worker
        return wrapper
    else:
        @wraps(coro_worker)
        async def wrapper(*args, **kwargs) -> Any:
            from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import RunCoro
            i: Interface = current_interface()
            return await i(RunCoro, cs_coro(coro_worker), *args, **kwargs)
        
        wrapper: coro_worker
        return wrapper


def coro_interfaces_arg_manager(event_loop, coro_scheduler: CoroSchedulerType):
    acf = ArgsManager(
        EArgs(event_loop, coro_scheduler, CoroType.auto),
    ).callable(await_coro_fast)
    atf = ArgsManager(
        EArgs(event_loop, coro_scheduler),
    ).callable(await_task_fast)
    
    return EArgs(await_coro_fast=acf, await_task_fast=atf)


class RunSchedulerInAsyncioLoop:
    def __init__(self, scheduler: CoroSchedulerType, idle_time: Optional[Union[int, float]]=None, loop: Optional[asyncio.AbstractEventLoop]=None, execute_every_X_iterations: int = 1):
        self.scheduler = scheduler
        self.scheduler.on_woke_up_callback = self.on_woke_up_callback
        if idle_time is None:
            idle_time = get_usable_min_sleep_interval()
        
        self.idle_time = get_min_sleep_interval() * (idle_time // get_min_sleep_interval())
        self.min_sleep_interval = max(get_min_sleep_interval(), self.idle_time or 0)
        self.usable_idle_time = (get_min_sleep_interval() * (idle_time // get_min_sleep_interval())) + get_countable_delta_time()
        self.loop = loop or asyncio.get_event_loop()
        self.handle = None  # type: Optional[asyncio.Handle]
        self.make_idle_when_possible = False
        self.need_to_stop_when_possible = False
        self.need_to_stop_now = False
        self.in_idle_state = False
        if execute_every_X_iterations < 1:
            execute_every_X_iterations = 1
        
        self.execute_every_X_iterations = execute_every_X_iterations
        self.current_iteration = execute_every_X_iterations
    
    def on_woke_up_callback(self):
        if self.in_idle_state:
            self.cancel_handle()
            self.in_idle_state = False
            self.register()

    def __call__(self, *args, **kwargs):
        self.current_iteration -= 1
        if self.current_iteration:
            self.register()
            return
        else:
            self.current_iteration = self.execute_every_X_iterations
        
        self.handle = None
        self.in_idle_state = False
        if self.need_to_stop_now:
            self.need_to_stop_now = False
            self.need_to_stop_when_possible = False
            return

        in_work = self.scheduler.iteration()
        
        if (not in_work) and self.need_to_stop_when_possible:
            self.need_to_stop_now = False
            self.need_to_stop_when_possible = False
            return

        idle_time = self.scheduler.next_event_after()
        # is_awake = self.scheduler.is_awake()
        is_idle = self.scheduler.is_idle()
        need_to_register = False
        if self.make_idle_when_possible:
            if not is_idle:
                need_to_register = True
        else:
            need_to_register = True

        if need_to_register:
            self.register()
        else:
            self.register_idle(idle_time)

    def ready_to_stop(self) -> bool:
        return not self.scheduler.in_work()

    def register(self):
        # self.register_idle()
        self.handle = self.loop.call_soon(self)

    def register_idle(self, idle_time: Optional[float] = None):
        if idle_time is None:
            idle_time = self.usable_idle_time
        
        # self.handle = self.loop.call_later(self.idle_time, self)
        if idle_time < self.min_sleep_interval:
            self.register()
        else:
            # self.handle = try_sleep(self.idle_time * (idle_time // self.idle_time), self.loop.call_later, self)
            # # self.handle = try_sleep(0.001, self.loop.call_later, self)
            self.in_idle_state = True
            idle_time = (self.idle_time * (idle_time // self.idle_time)) + get_countable_delta_time()
            self.handle = self.loop.call_later(idle_time, self)

    def cancel_handle(self):
        if self.handle and (not self.handle.cancelled()):
            self.handle.cancel()
            self.handle = None

    def stop(self):
        if self.handle and (not self.handle.cancelled()):
            self.need_to_stop_now = True
            self.handle.cancel()
            self.handle = None

    def stop_when_possible(self):
        self.need_to_stop_when_possible = True


# def call_soon_future(loop: Optional[asyncio.AbstractEventLoop], awaitable_coro_obj: Awaitable):
#     if __debug__: dlog('call_soon_future - start')
#     loop = loop or asyncio.get_event_loop()
#
#     def handler():
#         if __debug__: dlog('call_soon_future.handler - start')
#         loop.create_task(awaitable_coro_obj)
#         if __debug__: dlog('call_soon_future.handler - end')
#
#     loop.get_event_loop().call_soon(handler)
#     if __debug__: dlog('call_soon_future - end')


def call_soon_future(loop: Optional[asyncio.AbstractEventLoop], awaitable_coro_obj: Awaitable):
    if __debug__: dlog('call_soon_future - start')
    loop = loop or asyncio.get_event_loop()

    def handler():
        if __debug__: dlog('call_soon_future.handler - start')
        loop.create_task(awaitable_coro_obj)
        if __debug__: dlog('call_soon_future.handler - end')

    loop.get_event_loop().call_soon(handler)
    if __debug__: dlog('call_soon_future - end')


def create_task(loop: Optional[asyncio.AbstractEventLoop], acoro: Callable, *args, **kwargs) -> asyncio.Task:
    if __debug__: dlog('call_soon_future - start')
    loop = loop or asyncio.get_event_loop()
    awaitable_coro_obj = acoro(*args, **kwargs)
    return loop.create_task(awaitable_coro_obj)


def create_task_in_thread_pool(loop: Optional[asyncio.AbstractEventLoop], joiner: Optional[Awaitable[Any]], worker: Callable, *args, **kwargs):
    async def sync_handler(loop, joiner, worker, args, kwargs):
        worker_with_args = partial(worker, *args, **kwargs)
        result = await loop.run_in_executor(None, worker_with_args)
        if joiner is not None:
            await joiner(result)
                        
    create_task(loop, sync_handler, loop, joiner, worker, args, kwargs)


async def task_in_thread_pool(loop: Optional[asyncio.AbstractEventLoop], worker: Callable, *args, **kwargs):
    worker_with_args = partial(worker, *args, **kwargs)
    return await loop.run_in_executor(None, worker_with_args)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ctypes
import ctypes.wintypes as wintypes
from cengal.code_flow_control.chained_flow.versions.v_1.chained_flow import chain_reader, ChainInternalResult, \
    ResultExistence

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


LPDWORD = ctypes.POINTER(wintypes.DWORD)
LPOVERLAPPED = wintypes.LPVOID
LPSECURITY_ATTRIBUTES = wintypes.LPVOID

GENERIC_READ = 0x80000000
GENERIC_WRITE = 0x40000000
GENERIC_EXECUTE = 0x20000000
GENERIC_ALL = 0x10000000

FILE_SHARE_DELETE = 0x00000004
FILE_SHARE_READ = 0x00000001
FILE_SHARE_WRITE = 0x00000002

CREATE_NEW = 1
CREATE_ALWAYS = 2
OPEN_EXISTING = 3
OPEN_ALWAYS = 4
TRUNCATE_EXISTING = 5

FILE_ATTRIBUTE_NORMAL = 0x00000080

INVALID_HANDLE_VALUE = -1
INVALID_HANDLE_VALUE_CTYPES = wintypes.HANDLE(INVALID_HANDLE_VALUE).value

NULL = 0
FALSE = wintypes.BOOL(0)
TRUE = wintypes.BOOL(1)

WIN_UNICODE_PATH_PREFIX = '\\\\?\\'
WUPP = WIN_UNICODE_PATH_PREFIX


def judge(ok_status, result):
    if ok_status:
        return ResultExistence(False, None)
    else:
        return ResultExistence(True, ctypes.WinError())


def verdict_exceptions_consolidator(context, block_results_criteria=None, close=False):
    with chain_reader(context, block_results_criteria, close):
        if not context:
            bad_blocks = context.get_bad_blocks()
            root_exception = None
            last_exception = None
            for block in reversed(bad_blocks):
                block_result = block[3][1]
                if type(block_result) != ChainInternalResult:
                    win_error = block_result[1].result
                    if root_exception is None:
                        root_exception = last_exception = win_error
                    else:
                        last_exception.__context__ = win_error
                        last_exception = win_error
            if root_exception is not None:
                raise root_exception
            else:
                context.raise_bad_blocks()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

from cengal.data_manipulation.conversion.reinterpret_cast_management.manager import BaseAutoDerivedObjWrapper
from cengal.entities.copyable import copy__impl, deepcopy__impl
from typing import Dict, Type, Callable, Any, Tuple, FrozenSet, Set
from copy import Error as CopyError, copy, deepcopy

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class DeepCopyWrapper(BaseAutoDerivedObjWrapper):
    def __init__(self):
        super().__init__()
        self._wrapping_required_per_type: Dict[Type, bool] = dict()
        self._methods_per_type: Dict[Type, Dict[str, Callable]] = dict()
    
    def wrapping_required(self, obj: Any, base_type: Type, fields: Tuple[str], planned_type_name: str) -> bool:
        try:
            return self._wrapping_required_per_type[base_type]
        except KeyError:
            pass

        required: bool = False
        try:
            deepcopy(obj)
        except CopyError:
            required = True
        
        self._wrapping_required_per_type[base_type] = required
        return required
    
    def methods(self, obj: Any, base_type: Type, fields: Tuple[str]) -> Dict[str, Callable]:
        try:
            return self._methods_per_type[base_type]
        except KeyError:
            pass
        
        methods = {
            '__deepcopy__': deepcopy__impl,
        }
        self._methods_per_type[base_type] = methods
        return methods



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['WatchdogTimeoutError', 'Watchdog']


from cengal.parallel_execution.coroutines.coro_scheduler import *
import signal
from typing import Tuple, Optional
from threading import current_thread, main_thread, Thread
from cengal.time_management.sleep_tools import sleep
from cengal.math.numbers import RationalNumber
from signal import SIG_IGN, SIGUSR1, raise_signal, signal
from cengal.system import OS_API_TYPE


class WatchdogTimeoutError(Exception):
    pass


class HungedCoroInfo:
    def __init__(self, interface: Interface, coro_start_time: RationalNumber):
        self.interface: Interface = interface
        self.coro_start_time: RationalNumber = coro_start_time


class Watchdog(TypedService[None]):
    def __init__(self, loop: CoroSchedulerType):
        super().__init__(loop)
        self.handler_set: bool = False
        self.previous_handler = None
        self.keyboard_interrupt_emited: bool = False
        self.period: RationalNumber = 5
        if 'nt' == OS_API_TYPE:
            self.raised_signal = SIG_IGN
        else:
            self.raised_signal = SIGUSR1
        
        self.hunged_coro: Optional[HungedCoroInfo] = None
        self.watchdog_thread: Optional[Thread] = None
        self.watchdog_thread_name: str = '_CengalCoroutinesWatchdogDeamon_'
        self.watchdog_thread_allowed: bool = True
        self.watchdog_thread_in_work: bool = False

    def watchdog_thread_func(self):
        while self.watchdog_thread_allowed and (not self._loop._destroyed):
            sleep(self.period)
            self.watchdog_thread_in_work = True
            try:
                self.check_idle_coro()
            finally:
                self.watchdog_thread_in_work = False
    
    def check_idle_coro(self):
        loop: CoroSchedulerType = self._loop
        if get_current_coro_scheduler() is not loop:
            return
        
        curr_interface: Optional[Interface] = current_interface()
        if curr_interface is None:
            return

        if curr_interface.ignored_by_watchdog:
            return

        if loop.current_coro_start_time is None:
            return

        if self.hunged_coro is not None:
            return

        coro_execution_piece_delta_time = loop.get_coro_start_time() - loop.current_coro_start_time
        if coro_execution_piece_delta_time > self.period:
            raise_signal(self.raised_signal)
    
    def is_in_main_thread(self):
        return current_thread() is main_thread()

    def single_task_registration_or_immediate_processing(
            self, period: RationalNumber) -> Tuple[bool, None, None]:
        result = False
        if self.is_in_main_thread():
            self._loop.set_coro_time_measurement(True)
            if not self.handler_set:
                self.previous_handler = signal.signal(self.raised_signal, self.interrupt_handler)
                self.handler_set = True
            
            if self.watchdog_thread is None:
                self.watchdog_thread = Thread(target=self.watchdog_thread_func, name=self.watchdog_thread_name, daemon=True)
                self.watchdog_thread.start()

            self.period = period
            result = True
        
        return result, None, None

    def full_processing_iteration(self):
        self.make_dead()

    def in_work(self) -> bool:
        result: bool = self.keyboard_interrupt_emited
        return self.thrifty_in_work(result)
    
    def interrupt_handler(self, sig, frame):
        if self.hunged_coro is None:
            return
        
        loop: CoroSchedulerType = self._loop
        if get_current_coro_scheduler() is not loop:
            return
        
        if current_interface() is not self.hunged_coro.interface:
            return

        if loop.current_coro_start_time != self.hunged_coro.coro_start_time:
            return

        try:
            raise WatchdogTimeoutError
        finally:
            self.hunged_coro = None
    
    def destroy(self):
        if self.watchdog_thread is not None:
            self.watchdog_thread_allowed = False
            while self.watchdog_thread_in_work:
                sleep(0.0001, high_cpu_utilisation_mode=True)
            
            # self.watchdog_thread.join()  # we don ot need this since it is a daemon thread
            self.watchdog_thread = None

        if self.handler_set:
            signal.signal(self.raised_signal, self.previous_handler)
            self.handler_set = False



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'aterminate_tasks_explicit',
    'aterminate_tasks_implicit',
    'aterminate_tasks',
    'aterminate_tasks_im',
    'terminate_tasks_explicit',
    'terminate_tasks_implicit',
    'terminate_tasks',
    'terminate_tasks_im',
]


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.parallel_execution.coroutines.coro_scheduler import Interface, current_interface
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import Task
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro_list import PutCoroList, PSCP
from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import RunCoro
from cengal.parallel_execution.coroutines.coro_standard_services.kill_coro_list import KillCoroList, KillSingleCoroParams
from cengal.parallel_execution.coroutines.coro_tools.wait_tasks import GracefulTerminationSettings, \
    atask_graceful_destroyer, await_tasks_explicit

from typing import Optional, List, Callable


async def aterminate_tasks_explicit(
        i: Interface, 
        tasks: List[Task], 
        tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        wait_for_termination: bool = False, 
        ) -> bool:
    if not tasks:
        return False
    
    if graceful_termination_settings is None:
        termination_tasks_params: List[KillSingleCoroParams] = list([KillSingleCoroParams(task.coro_id, tree) for task in tasks if not task.done])
        if not termination_tasks_params:
            return False
        
        i(KillCoroList, termination_tasks)
    else:
        termination_tasks_params: List[PSCP] = [PSCP(atask_graceful_destroyer, task, tree, graceful_termination_settings) for task in tasks if not task.done]
        if not termination_tasks_params:
            return False
        
        termination_tasks: List[Task] = await i(PutCoroList, termination_tasks_params, True)
        if wait_for_termination:
            await await_tasks_explicit(i, termination_tasks, False, None, None, False, False, False, False)
    
    return True


async def aterminate_tasks_implicit(
        tasks: List[Task], 
        tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        wait_for_termination: bool = False, 
        ) -> bool:
    i: Interface = current_interface()
    return await aterminate_tasks_explicit(
        i, 
        tasks, 
        tree, 
        graceful_termination_settings, 
        wait_for_termination, 
        )


aterminate_tasks: Callable = aterminate_tasks_explicit
aterminate_tasks_im: Callable = aterminate_tasks_implicit


def terminate_tasks_explicit(
        i: Interface, 
        tasks: List[Task], 
        tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        wait_for_termination: bool = False, 
        ) -> bool:
    i(RunCoro, aterminate_tasks_explicit,
        tasks, 
        tree, 
        graceful_termination_settings, 
        wait_for_termination, 
        )


def terminate_tasks_implicit(
        tasks: List[Task], 
        tree: bool = False,
        graceful_termination_settings: Optional[GracefulTerminationSettings] = None,
        wait_for_termination: bool = False, 
        ) -> bool:
    i: Interface = current_interface()
    return terminate_tasks_explicit(
        i, 
        tasks, 
        tree, 
        graceful_termination_settings, 
        wait_for_termination, 
        )


terminate_tasks: Callable = terminate_tasks_explicit
terminate_tasks_im: Callable = terminate_tasks_implicit



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"

from cengal.code_flow_control.multiinterface_essence import *

"""

"""


Kilogram = float


class Point:
    def __init__(self, x: float, y: float):
        self.x: float = x
        self.y: float = y


class Vector:
    def __init__(self, start: Point, end: Point):
        self.start = start
        self.end = end


class Weight(EssenceModel):
    def __init__(self, weight: Kilogram):
        super(Weight, self).__init__()
        self.weight: Kilogram = weight


class MaterialPoint(EssenceModel):
    def __init__(self, material_point: Point):
        super(MaterialPoint, self).__init__()
        self.material_point = material_point


class Riding(EssenceModel):
    def __init__(self, riding_vector: Vector):
        super(Riding, self).__init__()
        self.riding_vector = riding_vector


class Flying(EssenceModel):
    def __init__(self, flying_vector: Vector):
        super(Flying, self).__init__()
        self.flying_vector = flying_vector


class Sitting(EssenceModel):
    def __init__(self):
        super(Sitting, self).__init__()
        self.sitter = None


class EssenceOnCasters(EssenceModel):
    emi_compatible_injectable_essence_model_classes: Set[Type['EssenceModel']] = {Riding}
    def __init__(self, dimension: Point):
        super(EssenceOnCasters, self).__init__()
        self.dimension: Point = dimension


class Chair(EssenceModel):
    def __init__(self, dimension: Point):
        super(Chair, self).__init__()
        self.dimension: Point = dimension


class ChairOnCasters(EssenceModel):
    def __init__(self, dimension: Point):
        super(ChairOnCasters, self).__init__()
        self.dimension: Point = dimension



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = [
    'CoroPriority', 'ThisCoroWasRequestedToBeKilled', 'LoopYieldPrioritySchedulerRequest', 'LoopYieldManagedBase', 'LoopYieldManaged',
    'LoopYieldManagedAsync', 'FakeLoopYieldManaged', 'LoopYieldPriorityScheduler', 'get_loop_yield',
    'gly', 'aget_loop_yield', 'agly', 'LoopYieldManagedAsyncExternal', 'FakeLoopYieldManagedAsync', 'external_aget_loop_yield', 'eagly'
]

from asyncio.events import Handle
from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_tools.await_coro import *
from enum import Enum
from typing import Callable, Dict, Tuple, Union, Type, Optional, Any, Set, Hashable
from cengal.time_management.repeat_for_a_time import Tracer, TimeLimitIsTooSmall
from cengal.code_flow_control.smart_values.versions.v_1 import ValueExistence
from async_generator import asynccontextmanager, async_generator, yield_
import asyncio
import sys


MIN_TIME = 0.000001


class CoroPriority(Enum):
    high = 0
    normal = 1
    low = 2


class LoopYieldPrioritySchedulerRequest(ServiceRequest):
    def register(self, default_priority: CoroPriority) -> 'LoopYieldManagedBase':
        return self._save(0, default_priority)

    def setup(self, max_delay: float) -> None:
        return self._save(1, max_delay)

    def change_priority(self, new_priority: CoroPriority, old_priority: CoroPriority) -> None:
        return self._save(2, new_priority, old_priority)

    def get(self) -> 'LoopYieldManagedBase':
        return self._save(3)

    def register_external(self, asyncio_loop: asyncio.AbstractEventLoop, default_priority: CoroPriority) -> 'LoopYieldManagedAsyncExternal':
        return self._save(4, asyncio_loop, default_priority)

    def register_external_asyncio_task(self, asyncio_loop: asyncio.AbstractEventLoop, task: asyncio.Task, default_priority: CoroPriority) -> 'LoopYieldManagedAsyncExternal':
        return self._save(5, asyncio_loop, task, default_priority)

    def change_priority_external(self, loop_yield: 'LoopYieldManagedAsyncExternal', new_priority: CoroPriority, old_priority: CoroPriority) -> None:
        return self._save(6, loop_yield, new_priority, old_priority)

    def del_external(self, loop_yield: 'LoopYieldManagedAsyncExternal') -> None:
        return self._save(7, loop_yield)

    def request_coro_kill(self, coro_id: CoroID) -> None:
        """Make request for a coro kill asynchronously (immidiately returns). An exception ThisCoroWasRequestedToBeKilled will be emmited in coro with a requested coro_id during one of it's next requests to LoopYield service

        Args:
            coro_id (CoroID): _description_

        Returns:
            ServiceRequest: _description_
        """        
        return self._save(8, coro_id)
    
    def kill_coro(self, coro_id: CoroID) -> None:
        """Make request for a coro kill synchronously (waits for next steps). An exception ThisCoroWasRequestedToBeKilled will be emmited in coro with a requested coro_id during one of it's next requests to LoopYield service. After an exception was sent to the requested coro_id, current coro will receive an answer from the service (None).

        Args:
            coro_id (CoroID): _description_

        Returns:
            ServiceRequest: _description_
        """        
        return self._save(9, coro_id)


class LoopYieldManagedBase:
    def __init__(self, interface: Interface, time_atom: ValueExistence,
                 default_priority: CoroPriority, service: Type[Service]):
        self._interface = None
        self.interface = interface
        self.time_atom = time_atom
        self.default_priority = default_priority
        self.priority = self.default_priority
        self.service = service
        self.tracer = None

    @property
    def interface(self):
        if self._interface is None:
            self._interface = current_interface()
        
        return self._interface

    @interface.setter
    def interface(self, value):
        self._interface = value


class LoopYieldManaged(LoopYieldManagedBase):
    def __init__(self, interface: Interface, time_atom: ValueExistence,
                 default_priority: CoroPriority, service: Type[Service]):
        super(LoopYieldManaged, self).__init__(interface, time_atom, default_priority, service)

    def __call__(self, priority: Optional[CoroPriority] = None):
        if priority is None:
            priority = self.default_priority
        
        if priority != self.priority:
            self.interface = None
            self.interface(self.service, LoopYieldPrioritySchedulerRequest().change_priority(priority,
                                                                                             self.priority))
            self.priority = priority
            try:
                self.tracer = Tracer(self.time_atom.result)
            except TimeLimitIsTooSmall as ex:
                self.tracer = Tracer(ex.min_time if ex.min_time is not None else MIN_TIME)
        
        if self.tracer is None:
            exception = None
            try:
                self.tracer = Tracer(self.time_atom.result)
            except TimeLimitIsTooSmall as ex:
                exception = ex

            if exception is not None:
                try:
                    self.tracer = Tracer(exception.min_time if exception.min_time is not None else MIN_TIME)
                except TimeLimitIsTooSmall as ex:
                    print(ex)
                
        if self.tracer is not None:
            if not self.tracer.iter():
                self.tracer = None
                self.interface = None
                self.interface(self.service)


class LoopYieldManagedAsync(LoopYieldManagedBase):
    def __init__(self, interface: Interface, time_atom: ValueExistence,
                 default_priority: CoroPriority, service: Type[Service]):
        super(LoopYieldManagedAsync, self).__init__(interface, time_atom, default_priority, service)

    async def __call__(self, priority: Optional[CoroPriority] = None):
        if priority is None:
            priority = self.default_priority
        if priority != self.priority:
            self.interface = None
            await self.interface(self.service, LoopYieldPrioritySchedulerRequest().change_priority(priority,
                                                                                                   self.priority))
            self.priority = priority
            try:
                self.tracer = Tracer(self.time_atom.result)
            except TimeLimitIsTooSmall as ex:
                self.tracer = Tracer(ex.min_time if ex.min_time is not None else MIN_TIME)
        if self.tracer is None:
            try:
                self.tracer = Tracer(self.time_atom.result)
            except TimeLimitIsTooSmall as ex:
                self.tracer = Tracer(ex.min_time if ex.min_time is not None else MIN_TIME)
        if not self.tracer.iter():
            self.tracer = None
            self.interface = None
            await self.interface(self.service)


class LoopYieldManagedAsyncExternal(LoopYieldManagedBase):
    def __init__(self, task_id: int, time_atom: ValueExistence,
                 default_priority: CoroPriority, service: Type[Service], coro_scheduler: CoroSchedulerType, asyncio_loop: asyncio.AbstractEventLoop):
        super(LoopYieldManagedAsyncExternal, self).__init__(None, time_atom, default_priority, service)
        self.task_id = task_id
        self.coro_scheduler = coro_scheduler
        self.asyncio_loop = asyncio_loop
        self.asyncio_task: asyncio.Task = None
        self.on_done_asyncio_coro: Callable = None

    async def __call__(self, priority: Optional[CoroPriority] = None):
        if priority is None:
            priority = self.default_priority
        if priority != self.priority:
            await await_task_fast(self.asyncio_loop, self.coro_scheduler, self.service,
                                  LoopYieldPrioritySchedulerRequest().change_priority_external(self, priority, self.priority))
            self.priority = priority
            try:
                self.tracer = Tracer(self.time_atom.result)
            except TimeLimitIsTooSmall as ex:
                self.tracer = Tracer(ex.min_time if ex.min_time is not None else MIN_TIME)
        if self.tracer is None:
            try:
                self.tracer = Tracer(self.time_atom.result)
            except TimeLimitIsTooSmall as ex:
                self.tracer = Tracer(ex.min_time if ex.min_time is not None else MIN_TIME)
        if not self.tracer.iter():
            self.tracer = None
            await await_task_fast(self.asyncio_loop, self.coro_scheduler, self.service)


class FakeLoopYieldManaged:
    def __call__(self, priority: Optional[CoroPriority] = None):
        pass

class FakeLoopYieldManagedAsync:
    async def __call__(self, priority: Optional[CoroPriority] = None):
        pass


class ThisCoroWasRequestedToBeKilled(Exception):
    pass


class LoopYieldPriorityScheduler(TypedService[None], EntityStatsMixin):
    def __init__(self, loop: CoroSchedulerType):
        super(LoopYieldPriorityScheduler, self).__init__(loop)

        # loop.add_global_on_coro_del_handler(self._on_coro_del_handler_global)  # Todo: switch to local coro del handler

        self._request_workers = {
            0: self._on_register,
            1: self._on_setup,
            2: self._on_change_priority,
            3: self._on_get,
            4: self._on_register_external,
            5: self._on_register_external_asyncio_task,
            6: self._on_change_priority_external,
            7: self._on_del_external,
            8: self._request_coro_kill,
            9: self._kill_coro,
        }

        self.sigma = {
            0: 0.6827,
            1: 0.9545 - 0.6827,
            2: 1.0 - 0.9545,
        }

        self.max_delay = 0.01
        self.max_delays = {
            0: 0.0,
            1: 0.0,
            2: 0.0,
        }
        self.compute_delays()

        self.all_yield_objects = dict()  # type: Dict[CoroID, LoopYieldManagedBase]
        self.task_counter = Counter()
        while self.task_counter.get() <= 0:
            pass
        
        self.yields_num: int = 0
        self.coroutines_requested_to_be_deleted: Set[CoroID] = set()
        self.coroutines_requested_to_be_deleted_by_waiters: Dict[CoroID, Set[CoroID]] = dict()
        self.finished_waiters_for_coro_kill: Set[CoroID] = set()
        self.asyncio_task_ids: Dict[Hashable, int] = dict()

        self.yields_by_priority: Dict[CoroPriority, int] = {
            CoroPriority.high:   0,
            CoroPriority.normal: 0,
            CoroPriority.low:    0,
        }

        self.time_atom_by_priority = {
            CoroPriority.high:   ValueExistence(True, self.max_delays[0]),
            CoroPriority.normal: ValueExistence(True, self.max_delays[1]),
            CoroPriority.low:    ValueExistence(True, self.max_delays[2]),
        }

    def get_entity_stats(self, stats_level: 'EntityStatsMixin.StatsLevel' = EntityStatsMixin.StatsLevel.debug) -> Tuple[str, Dict[str, Any]]:
        coroutines_requested_to_be_killed = self.coroutines_requested_to_be_deleted | set(self.coroutines_requested_to_be_deleted_by_waiters.keys())
        return type(self).__name__, {
            'task counter': self.task_counter._index,
            'yields num': self.yields_num,
            'max_delay': self.max_delay,
            'max_delays': self.max_delays,
            'affected coroutines num': len(self.all_yield_objects),
            'coroutines num by priority': {
                'high': self.yields_by_priority[CoroPriority.high],
                'normal': self.yields_by_priority[CoroPriority.normal],
                'low': self.yields_by_priority[CoroPriority.low],
            },
            'time atoms by priority': {
                'high': self.time_atom_by_priority[CoroPriority.high].result,
                'normal': self.time_atom_by_priority[CoroPriority.normal].result,
                'low': self.time_atom_by_priority[CoroPriority.low].result,
            },
            'coroutines requested to be killed': {
                'num': len(coroutines_requested_to_be_killed),
                'list': coroutines_requested_to_be_killed,
            }
        }

    def single_task_registration_or_immediate_processing(self, request: Optional[LoopYieldPrioritySchedulerRequest]=None
                                                         ) -> Tuple[bool, Any, None]:
        self.yields_num += 1
        coro_should_be_killed = False
        coro_id = self.current_caller_coro_info.coro_id
        if coro_id in self.coroutines_requested_to_be_deleted:
            coro_should_be_killed = True
        
        if coro_id in self.coroutines_requested_to_be_deleted_by_waiters:
            self.finished_waiters_for_coro_kill |= self.coroutines_requested_to_be_deleted_by_waiters[coro_id]
            self.make_live()
            coro_should_be_killed = True
        
        if request is not None:
            if coro_should_be_killed and (self._on_del_external != self._request_workers[request.request_type]):
                return True, None, ThisCoroWasRequestedToBeKilled
                
            return self.resolve_request(request)

        if coro_should_be_killed:
            return True, None, ThisCoroWasRequestedToBeKilled
        
        return True, None, None

    def full_processing_iteration(self):
        for coro_id in self.finished_waiters_for_coro_kill:
            self.register_response(coro_id, None)
        
        self.compute_time_atoms()

    def compute_time_atoms(self):
        if self.all_yield_objects:
            top_sigma = 0

            if self.yields_by_priority[CoroPriority.high]:
                top_sigma += 1

            if self.yields_by_priority[CoroPriority.normal]:
                top_sigma += 1

            if self.yields_by_priority[CoroPriority.low]:
                top_sigma += 1

            median_time_atom = 1 / len(self.all_yield_objects)  # !!! Possible division by zero! Conditional must not be removed!
            if 1 == top_sigma:
                sigma_0_time_atom = min(median_time_atom, self.max_delays[0])
                self.time_atom_by_priority[CoroPriority.high].result = sigma_0_time_atom
                self.time_atom_by_priority[CoroPriority.normal].result = sigma_0_time_atom
                self.time_atom_by_priority[CoroPriority.low].result = sigma_0_time_atom
            elif 2 == top_sigma:
                sigma_0_time_atom = min(median_time_atom * self.sigma[0], self.max_delays[0])
                sigma_1_time_atom = min(median_time_atom * self.sigma[1], self.max_delays[1])
                if self.yields_by_priority[CoroPriority.high]:
                    self.time_atom_by_priority[CoroPriority.high].result = sigma_0_time_atom
                    self.time_atom_by_priority[CoroPriority.normal].result = sigma_1_time_atom
                    self.time_atom_by_priority[CoroPriority.low].result = sigma_1_time_atom
                else:
                    self.time_atom_by_priority[CoroPriority.high].result = sigma_0_time_atom
                    self.time_atom_by_priority[CoroPriority.normal].result = sigma_0_time_atom
                    self.time_atom_by_priority[CoroPriority.low].result = sigma_1_time_atom
            elif 3 == top_sigma:
                sigma_0_time_atom = min(median_time_atom * self.sigma[0], self.max_delays[0])
                sigma_1_time_atom = min(median_time_atom * self.sigma[1], self.max_delays[1])
                sigma_2_time_atom = min(median_time_atom * self.sigma[2], self.max_delays[2])
                self.time_atom_by_priority[CoroPriority.high].result = sigma_0_time_atom
                self.time_atom_by_priority[CoroPriority.normal].result = sigma_1_time_atom
                self.time_atom_by_priority[CoroPriority.low].result = sigma_2_time_atom
        
        self.make_dead()

    def in_work(self) -> bool:
        result: bool = bool(self.finished_waiters_for_coro_kill) or bool(self.all_yield_objects)
        return self.thrifty_in_work(result)

    def compute_delays(self):
        self.max_delays = {
            0: self.max_delay * self.sigma[0],
            1: self.max_delay * self.sigma[1],
            2: self.max_delay * self.sigma[2],
        }

    def _on_register(self, default_priority: CoroPriority):
        task_id = self.current_caller_coro_info.coro_id
        if task_id in self.all_yield_objects:
            loop_yield = self.all_yield_objects[task_id]
        else:
            interface: Interface = self.current_caller_coro_info.coro.interface
            if isinstance(interface, InterfaceGreenlet):
                loop_yield = LoopYieldManaged(interface,
                                            self.time_atom_by_priority[default_priority],
                                            default_priority,
                                            type(self))
            elif isinstance(interface, InterfaceAsyncAwait):
                loop_yield = LoopYieldManagedAsync(interface,
                                                self.time_atom_by_priority[default_priority],
                                                default_priority,
                                                type(self))
            else:
                raise NotImplementedError
            
            loop_yield.time_atom = self.time_atom_by_priority[loop_yield.priority]
            self.all_yield_objects[interface.coro_id] = loop_yield
            self.current_caller_coro_info.coro.add_on_coro_del_handler(self._on_coro_del_handler)
            self.yields_by_priority[loop_yield.priority] += 1
            self.make_live()
        
        return True, loop_yield, None

    def _on_setup(self, max_delay: float):
        self.max_delay = max_delay
        self.compute_delays()
        # self.compute_time_atoms()
        self.make_live()
        return True, None, None

    def _on_change_priority(self, new_priority: CoroPriority, old_priority: CoroPriority):
        self.yields_by_priority[old_priority] -= 1
        self.yields_by_priority[new_priority] += 1
        loop_yield = self.all_yield_objects[self.current_caller_coro_info.coro_id]
        loop_yield.time_atom = self.time_atom_by_priority[new_priority]
        self.make_live()
        return True, None, None

    def _on_get(self):
        return True, self.all_yield_objects.get(self.current_caller_coro_info.coro_id), None

    def get_yield_object(self, coro_id: CoroID) -> LoopYieldManagedBase:
        return self.all_yield_objects.get(coro_id)
    
    def _on_register_external(self, asyncio_loop: asyncio.AbstractEventLoop, default_priority: CoroPriority):
        task_id = -(self.task_counter.get())
        loop_yield = LoopYieldManagedAsyncExternal(task_id,
                                                   self.time_atom_by_priority[default_priority],
                                                   default_priority,
                                                   type(self),
                                                   self._loop,
                                                   asyncio_loop)
        loop_yield.time_atom = self.time_atom_by_priority[loop_yield.priority]
        self.all_yield_objects[task_id] = loop_yield
        self.yields_by_priority[loop_yield.priority] += 1
        self.make_live()
        return True, loop_yield, None
    
    def _on_register_external_asyncio_task(self, asyncio_loop: asyncio.AbstractEventLoop, task: asyncio.Task, default_priority: CoroPriority):
        asyncio_task_id = id(task)
        if asyncio_task_id not in self.asyncio_task_ids:
            self.asyncio_task_ids[asyncio_task_id] = -(self.task_counter.get())
            
        task_id = self.asyncio_task_ids[asyncio_task_id]
        if task_id in self.all_yield_objects:
            loop_yield = self.all_yield_objects[task_id]
        else:
            loop_yield = LoopYieldManagedAsyncExternal(task_id,
                                                    self.time_atom_by_priority[default_priority],
                                                    default_priority,
                                                    type(self),
                                                    self._loop,
                                                    asyncio_loop)
            def on_done_asyncio_coro(future):
                self._on_del_external(loop_yield)
                task.remove_done_callback(loop_yield.on_done_asyncio_coro)
            
            loop_yield.asyncio_task = task
            loop_yield.on_done_asyncio_coro = on_done_asyncio_coro
            task.add_done_callback(loop_yield.on_done_asyncio_coro)
            loop_yield.time_atom = self.time_atom_by_priority[loop_yield.priority]
            self.all_yield_objects[task_id] = loop_yield
            self.yields_by_priority[loop_yield.priority] += 1
            self.make_live()
        
        return True, loop_yield, None
    
    def _on_change_priority_external(self, loop_yield: LoopYieldManagedAsyncExternal, new_priority: CoroPriority, old_priority: CoroPriority):
        self.yields_by_priority[old_priority] -= 1
        self.yields_by_priority[new_priority] += 1
        loop_yield.time_atom = self.time_atom_by_priority[new_priority]
        self.make_live()
        return True, None, None
    
    def _on_del_external(self, loop_yield: LoopYieldManagedAsyncExternal):
        task_id = loop_yield.task_id
        if task_id in self.all_yield_objects:
            priority = self.all_yield_objects[task_id].priority
            del self.all_yield_objects[task_id]
            self.yields_by_priority[priority] -= 1
        self.make_live()
        return True, None, None
    
    def _request_coro_kill(self, coro_id: CoroID) -> ServiceProcessingResponse:
        self.coroutines_requested_to_be_deleted.add(coro_id)
        return True, None, None
    
    def _kill_coro(self, coro_id: CoroID) -> ServiceProcessingResponse:
        waiter_coro_id = self.current_caller_coro_info
        if coro_id not in self.coroutines_requested_to_be_deleted_by_waiters:
            self.coroutines_requested_to_be_deleted_by_waiters[coro_id] = set()
        
        self.coroutines_requested_to_be_deleted_by_waiters[coro_id].add(waiter_coro_id)
        return False, None, None

    def _on_coro_del_handler_global(self, coro: CoroWrapperBase) -> bool:
        coro_id = coro.coro_id
        if coro_id in self.all_yield_objects:
            priority = self.all_yield_objects[coro_id].priority
            del self.all_yield_objects[coro_id]
            self.yields_by_priority[priority] -= 1
            self.make_live()
        return False

    def _on_coro_del_handler(self, coro: CoroWrapperBase) -> bool:
        coro_id = coro.coro_id
        priority = self.all_yield_objects[coro_id].priority
        del self.all_yield_objects[coro_id]
        self.yields_by_priority[priority] -= 1
        self.make_live()
        return False


LoopYieldPrioritySchedulerRequest.default_service_type = LoopYieldPriorityScheduler


def get_loop_yield(default_priority: CoroPriority = CoroPriority.normal) -> Union[LoopYieldManaged, FakeLoopYieldManaged]:
    loop = CoroScheduler.current_loop()
    if loop is None:
        return FakeLoopYieldManaged()  # running not from inside the loop

    interface = loop.current_interface()
    if interface is None:
        return FakeLoopYieldManaged()  # running from Service

    # ly = interface(LoopYieldPriorityScheduler, LoopYieldPrioritySchedulerRequest().get())
    ly = interface._loop.get_service_instance(LoopYieldPriorityScheduler).get_yield_object(interface.coro_id)
    if ly is None:
        ly = interface(LoopYieldPriorityScheduler, LoopYieldPrioritySchedulerRequest().register(default_priority))
    
    return ly


gly = get_loop_yield


async def aget_loop_yield(default_priority: CoroPriority = CoroPriority.normal) -> Union[LoopYieldManagedAsync, FakeLoopYieldManagedAsync]:
    loop = CoroScheduler.current_loop()
    if loop is None:
        return FakeLoopYieldManagedAsync()  # running not from inside the loop

    interface = loop.current_interface()
    if interface is None:
        return FakeLoopYieldManagedAsync()  # running from Service

    # ly = await interface(LoopYieldPriorityScheduler, LoopYieldPrioritySchedulerRequest().get())
    ly = interface._loop.get_service_instance(LoopYieldPriorityScheduler).get_yield_object(interface.coro_id)
    if ly is None:
        ly = await interface(LoopYieldPriorityScheduler, LoopYieldPrioritySchedulerRequest().register(default_priority))
    
    return ly


agly = aget_loop_yield


@asynccontextmanager
@async_generator
async def external_aget_loop_yield(default_priority: CoroPriority = CoroPriority.normal, coro_scheduler: Optional[CoroSchedulerType]=None, asyncio_loop: Optional[asyncio.AbstractEventLoop]=None):
    if coro_scheduler is None:
        coro_scheduler = CoroScheduler.current_loop()
    
    if coro_scheduler is None:
        await yield_(FakeLoopYieldManagedAsync())  # can not determine coro scheduler loop
    else:
        if asyncio_loop is None:
            asyncio_loop = asyncio.get_event_loop()
        
        if (3, 7) <= sys.version_info:
            ly: LoopYieldManagedAsyncExternal = await await_task_fast(
                asyncio_loop, coro_scheduler, LoopYieldPriorityScheduler,
                LoopYieldPrioritySchedulerRequest().register_external_asyncio_task(asyncio_loop, asyncio.current_task(loop=asyncio_loop), default_priority))
        else:
            ly: LoopYieldManagedAsyncExternal = await await_task_fast(
                asyncio_loop, coro_scheduler, LoopYieldPriorityScheduler,
                LoopYieldPrioritySchedulerRequest().register_external(asyncio_loop, default_priority))

        try:
            await yield_(ly)
        finally:
            if (3, 7) > sys.version_info:
                await await_task_fast(asyncio_loop, coro_scheduler, LoopYieldPriorityScheduler,
                                    LoopYieldPrioritySchedulerRequest().del_external(ly))

eagly = external_aget_loop_yield



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['bit_cast__general', 'bit_cast__float_to_uint32', 'bit_cast__uint32_to_float']


from struct import pack, unpack
from typing import Any


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def bit_cast__general(data: Any, bit_cast_from: str, bit_cast_to: str) -> Any:
    """See https://docs.python.org/3/library/struct.html

    Args:
        data (Any): _description_
        bit_cast_from (str): see https://docs.python.org/3/library/struct.html#format-characters
        bit_cast_to (str): see https://docs.python.org/3/library/struct.html#format-characters

    Returns:
        Any: _description_
    """
    return unpack(bit_cast_to, pack(bit_cast_from, data))


def bit_cast__float_to_uint32(x: float) -> int:
    return unpack('I', pack('f', x))


def bit_cast__uint32_to_float(x: int) -> float:
    return unpack('f', pack('I', x))


def bit_cast__double_to_uint64(x: float) -> int:
    return unpack('Q', pack('d', x))


def bit_cast__uint64_to_double(x: int) -> float:
    return unpack('d', pack('Q', x))



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cengal.system import PLATFORM_NAME, PYTHON_VERSION, PYTHON_VERSION_INT
from cengal.text_processing.help_tools import get_text_in_brackets_offset
from cengal.system import PLATFORM_NAME
import binascii
import os, os.path
import pickle
import datetime
import base64
import json
from inspect import currentframe, getframeinfo, getouterframes
import traceback
import sys
import struct
from cengal.modules_management.alternative_import import alt_import

with alt_import('lzma') as lzma:
    if lzma is None:
        import lzmaffi.compat
        lzmaffi.compat.register()
        import lzma

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def join_path_example(a, b):
    """
    k = os.path.join('D:\\Dev\\Workspace\\My', 'XCOM Modding\\NewHitDamageCalc', 'Long War\\15f\\.hg\\store', '00changelog.i')
    k == 'D:\\Dev\\Workspace\\My\\XCOM Modding\\NewHitDamageCalc\\Long War\\15f\\.hg\\store\\00changelog.i'
    :param a:
    :param b:
    :return:
    """
    os.path.join(a, b)


def json_to_printable_string(json_obj):
    result = json.dumps(json_obj, sort_keys=True, indent=4, separators=(',', ': '))
    return result


def make_readable_json(json_string):
    result = json.dumps(json.loads(json_string), indent=4)
    return result


def print_exception_traceback_info():
    exception = sys.exc_info()
    formattedTraceback = traceback.format_exception(exception[0], exception[1], exception[2])
    exception = exception[:2] + (formattedTraceback,)
    trace = ''
    for line in exception[2]:
        trace += line
    print(trace, file=sys.stderr)
    print(exception[0])
    print(exception[1])


# def load_old_class_to_the_new_one(oldStyleObject, newStyleObject, copyAllUnnecessaryMethods=False):
#     old_style_members = members = [attr for attr in dir(oldStyleObject) if not callable(attr) and not attr.startswith("__")]
#     new_style_members = members = [attr for attr in dir(newStyleObject) if not callable(attr) and not attr.startswith("__")]
#
#     for method in old_style_members:
#         method_allowed = False
#         if copyAllUnnecessaryMethods:
#             method_allowed = True
#         else:
#             if method in new_style_members:
#                 method_allowed = True
#             else:
#                 method_allowed = False
#
#         if method_allowed:
#             data = getattr(oldStyleObject, method)
#             setattr(newStyleObject, method, data)

def load_old_class_to_the_new_one(source_object_of_old_class,
                                  destination_object_of_new_class,
                                  copy_all_unnecessary_methods=False,
                                  has_list_of_methods=False,
                                  ignore_protected_methods=False):
    new_style_members = None
    if not copy_all_unnecessary_methods:
        if has_list_of_methods:
            new_style_members = destination_object_of_new_class.previouslyKnownSetOfMethodsToCopyAnObject
        else:
            new_style_members = dir(destination_object_of_new_class)

    list_of_old_style_object_methods = None
    if has_list_of_methods:
        list_of_old_style_object_methods = source_object_of_old_class.previouslyKnownSetOfMethodsToCopyAnObject
    else:
        list_of_old_style_object_methods = dir(source_object_of_old_class)

    template = None
    template_size = 0
    if ignore_protected_methods:
        template = '_'
        template_size = 1
    else:
        template = '__'
        template_size = 2
    method_allowed = False
    for attr in list_of_old_style_object_methods:
        if has_list_of_methods:
            method_allowed = True
        else:
            if (not callable(attr)) and (template != attr[:template_size]):
                method_allowed = True

        if method_allowed:
            if copy_all_unnecessary_methods:
                method_allowed = True
            else:
                if attr in new_style_members:
                    method_allowed = True
                else:
                    method_allowed = False

        if method_allowed:
            data = getattr(source_object_of_old_class, attr)
            setattr(destination_object_of_new_class, attr, data)


def bisect_search(data, max_value):
    result = tuple()

    min_tags_qnt = 2
    needed_items_qnt = 100
    left_bound = 0
    previous_left_bound = left_bound
    right_bound = len(data) - 1
    previous_right_bound = right_bound

    if (right_bound - left_bound) >= 1:
        # find right boundary
        while (right_bound - left_bound) > 0:
            if data[right_bound] > max_value:
                right_bound = previous_right_bound
                break
            previous_right_bound = right_bound
            diff = round(abs(right_bound - left_bound) / 2)
            if diff < 1:
                diff = 1
            right_bound -= diff
            if right_bound < left_bound:
                right_bound = left_bound

        # find left boundary
        while (right_bound - left_bound) > 0:
            if data[left_bound] <= max_value:
                left_bound = previous_left_bound
                break
            previous_left_bound = left_bound
            diff = round(abs(right_bound - left_bound) / 2)
            if diff < 1:
                diff = 1
            left_bound += diff
            if left_bound > right_bound:
                left_bound = right_bound

    if (right_bound - left_bound) > 1:
        # recursive
        result = bisect_search(data[left_bound:right_bound], max_value)
    else:
        result = (left_bound, right_bound)

    return result


def current_line():
    (frame, filename, line_number, function_name, lines, index) = getouterframes(currentframe())[1]
    result = (filename, function_name, line_number, lines, index)
    return result


def isn(parameter, value):
    """
    usage: k = isn(k, 'default')
    alternative: if k is not None: k = 'default'
    :param parameter:
    :param value:
    :return:
    """
    # if parameter is None:
    #     return value
    # else:
    #     return parameter
    return parameter or value


def parse_cmd_input_dir_list(search_dir_list_raw):
    """
    :param search_dir_list_raw: search_dir_list_raw = input('Choose a search dir (S:\\ome\\Dir\\Name) or
        (["S:\\ome\\Dir"], ["D:\\ir"], ["K:"]): ')
    :return:
    """
    search_dir_list = list()
    if '"' in search_dir_list_raw:
        while '"' in search_dir_list_raw:
            result = get_text_in_brackets_offset(search_dir_list_raw, '["', '"]')
            another_dir_name = result[0].strip()
            offset = result[1]
            if len(another_dir_name) > 0:
                search_dir_list.append(another_dir_name)
            search_dir_list_raw = search_dir_list_raw[offset:]
    else:
        search_dir_list.append(search_dir_list_raw.strip())
    return search_dir_list


def check_int_value(value, value_size=4):
    hi_value = 0
    lo_value = 0
    if value_size == 1:
        hi_value = 127
        lo_value = -128
    elif value_size == 2:
        hi_value = 32767
        lo_value = -32768
    elif value_size == 3:
        hi_value = 8388607
        lo_value = -8388608
    elif value_size == 4:
        hi_value = 2147483647
        lo_value = -2147483648
    if value < lo_value:
        value = lo_value
    if value > hi_value:
        value = hi_value
    return value





#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from typing import Callable, List
from types import CodeType
from cengal.code_flow_control.python_bytecode_manipulator import *
from cengal.introspection.inspect import is_async, is_callable


def gly_patch(entity: Callable) -> CodeType:
    # TODO: try to complise strings using _get_code_object() instead of get_code()
    entity_code_type: CodeTypeEnum = code_type(entity)
    if entity_code_type is None:
        raise RuntimeError('Entity {entity} cann not be patched')
    
    fn_code = get_code(entity)
    op_sequence, labels, op_by_label = OpSequence.from_bytecode_sequence(fn_code.co_code)
    op_sequence: OpSequence = op_sequence
    co_consts = list(fn_code.co_consts)
    first_new_const_index: int = len(co_consts)
    co_names = list(fn_code.co_names)
    first_new_name_index: int = len(co_names)
    co_varnames = list(fn_code.co_varnames)
    first_new_varname_index: int = len(co_varnames)
    co_nlocals = fn_code.co_nlocals


    co_consts.extend((0, ('gly',)))
    co_names.extend(('cengal.parallel_execution.coroutines.coro_standard_services.loop_yield', 'gly'))
    co_varnames.extend(('gly', 'ly'))
    co_nlocals = len(co_varnames)

    initialization: List[Instruction] = [
        mi('LOAD_CONST', first_new_const_index + 0),
        mi('LOAD_CONST', first_new_const_index + 1),
        mi('IMPORT_NAME', first_new_name_index + 0),
        mi('IMPORT_FROM', first_new_name_index + 1),
        mi('STORE_FAST', first_new_varname_index + 0),
        mi('POP_TOP'),
        mi('LOAD_FAST', first_new_varname_index + 0),
        mi('CALL_FUNCTION', 0),
        mi('STORE_FAST', first_new_varname_index + 1),
    ]
    initialization_op_sequence, _, _ = OpSequence.from_instructions(initialization)
    initialization_op_sequence: OpSequence = initialization_op_sequence

    ly_call: List[Instruction] = [
        mi('LOAD_FAST', first_new_varname_index + 1),
        mi('CALL_FUNCTION', 0),
        mi('POP_TOP'),
    ]
    ly_call_op_sequence, _, _ = OpSequence.from_instructions(ly_call)
    ly_call_op_sequence: OpSequence = ly_call_op_sequence

    op_sequence.insert_op_sequence(0, initialization_op_sequence)

    # for
    FOR_ITER = opcode('FOR_ITER')

    index = len(initialization) - 1
    while True:
        index += 1
        if index >= len(op_sequence.op_sequence):
            break

        op, _, _, _, _, _ = op_sequence.op_sequence[index]
        if FOR_ITER == op:
            place_to_insert = index + 2
            op_sequence.insert_op_sequence(place_to_insert, ly_call_op_sequence)

    JUMP_ABSOLUTE = opcode('JUMP_ABSOLUTE')

    index = len(initialization) - 1
    while True:
        index += 1
        if index >= len(op_sequence.op_sequence):
            break

        op, _, _, _, _, _ = op_sequence.op_sequence[index]
        arg = op_sequence.get_arg(index)
        if JUMP_ABSOLUTE == op:
            place_to_insert = op_sequence.op_sequence_offset_map.new_by_original[arg_to_op_index(arg)]
            place_to_insert_op, place_to_insert_arg, _, _, _, _ = op_sequence.op_sequence[place_to_insert]
            if FOR_ITER == place_to_insert_op:
                continue

            op_sequence.insert_op_sequence(index, ly_call_op_sequence)
            index += len(ly_call_op_sequence)
    
    fix_labels(op_sequence, op_by_label)

    new_code_type = CodeType(
        fn_code.co_argcount,
        fn_code.co_posonlyargcount,
        fn_code.co_kwonlyargcount,
        co_nlocals,
        fn_code.co_stacksize,
        fn_code.co_flags,
        op_sequence.to_bytes(),
        tuple(co_consts),
        tuple(co_names),
        tuple(co_varnames),
        fn_code.co_filename,
        fn_code.co_name,
        fn_code.co_firstlineno,
        fn_code.co_lnotab,
        fn_code.co_freevars,
        fn_code.co_cellvars,
    )
    set_code(entity, new_code_type)
    return fn_code


def agly_patch(entity: Callable) -> Callable:
    # TODO: try to complise strings using _get_code_object() instead of get_code()
    entity_code_type: CodeTypeEnum = code_type(entity)
    if entity_code_type is None:
        raise RuntimeError( 'Entity {entity} cann not be patched')
    
    fn_code = get_code(entity)
    return fn_code


def gly_patched(func: Callable) -> Callable:
    """Example:
        @gly_patched
        def func():
            for i in range(200):
                print(i)
                for j in range(100):
                    print(j)
            
            i = 100
            while i > 0:
                print(i)
                i -= 1

    Is equivalent to:
        from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import gly
        def func():
            ly = gly()
            for i in range(200):
                print(i)
                for j in range(100):
                    print(j)
                    ly()
                
                ly()
            
            i = 100
            while i > 0:
                print(i)
                i -= 1
                ly()

    Args:
        func (Callable): _description_

    Returns:
        Callable: _description_
    """    
    gly_patch(func)
    return func


glyp = gly_patched
gp = gly_patched


def agly_patched(func: Callable) -> Callable:
    agly_patch(func)
    return func


aglyp = agly_patched
agp = agly_patched


class GlyPatchManagerError(Exception):
    pass


class GlyPatchManager:
    patched_functions = dict()
    call_tree = dict()

    def __call__(self, func: Callable, tree: bool = True) -> Callable:
        if func not in self.patched_functions:
            if tree:
                if is_async(func):
                    agly_patched_tree(func, self)
                elif is_callable(func):
                    gly_patched_tree(func, self)
                else:
                    raise GlyPatchManagerError(f'Entity {func} can not be patched: it must be a callable (either async or not)')
            else:
                if is_async(func):
                    fn_code = agly_patch(func)
                elif is_callable(func):
                    fn_code = gly_patch(func)
                else:
                    raise GlyPatchManagerError(f'Entity {func} can not be patched: it must be a callable (either async or not)')
            
                self.patched_functions[func] = fn_code
        
        return func

    def restore(self, func: Callable, tree: bool = True) -> Callable:
        if func in self.patched_functions:
            if tree:
                self._unpatch_tree(func)
            else:
                set_code(func, self.patched_functions[func])

        return func
    
    def _unpatch_tree(self, func: Callable) -> Callable:
        raise NotImplementedError


def gly_patched_tree(entity: Callable, gly_patch_manager: Optional[GlyPatchManager] = None) -> Callable:
    raise NotImplementedError


def agly_patched_tree(entity: Callable, gly_patch_manager: Optional[GlyPatchManager] = None) -> Callable:
    raise NotImplementedError



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['AppDirPath']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


# More info:
# https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
# https://wiki.archlinux.org/title/XDG_Base_Directory
# https://refspecs.linuxfoundation.org/FHS_3.0/fhs-3.0.html


import os
from cengal.file_system.directory_manager import ensure_dir as ensure_dir_exists
from .app_directory_types import AppDirectoryType
from .app_dir_path_base import AppDirPathBase, DirTypeMappingItem, DirTypeMappingItem, BaseDirID, DirNameOrPath, norm_dir_name_or_path
from typing import Optional, Tuple, Dict
from functools import lru_cache
from enum import Enum


class LinuxDirectoryType(Enum):
    local_data = 0
    local_cache = 1
    local_config = 2
    local_log = 3
    local_temp = 4
    local_runtime = 5

    program_data = 6
    program_cache = 7
    program_config = 8
    program_log = 9
    program_temp = 10
    program_runtime = 11

    program_files = 12

    user_profile_data = 13
    user_profile_program_files = 14
    user_profile_program_files_common = 15

    local_static_data = 16
    program_static_data = 17


dir_type_mapping: Dict[AppDirectoryType, Tuple[LinuxDirectoryType, Optional[DirNameOrPath], Optional[DirNameOrPath]]] = {
    AppDirectoryType.local_data: (LinuxDirectoryType.local_data, None, ('local', 'data')),
    AppDirectoryType.local_static_data: (LinuxDirectoryType.local_static_data, None, ('local', 'static_data')),
    AppDirectoryType.local_cache: (LinuxDirectoryType.local_cache, None, ('local', 'cache')),
    AppDirectoryType.local_config: (LinuxDirectoryType.local_config, None, ('local', 'config')),
    AppDirectoryType.local_log: (LinuxDirectoryType.local_log, None, ('local', 'log')),
    AppDirectoryType.local_temp: (LinuxDirectoryType.local_temp, None, ('local', 'temp')),
    AppDirectoryType.local_runtime: (LinuxDirectoryType.local_runtime, None, ('local', 'runtime')),

    AppDirectoryType.local_low_data: (LinuxDirectoryType.local_data, None, ('local_low', 'data')),
    AppDirectoryType.local_low_static_data: (LinuxDirectoryType.local_static_data, None, ('local_low', 'static_data')),
    AppDirectoryType.local_low_cache: (LinuxDirectoryType.local_cache, None, ('local_low', 'cache')),
    AppDirectoryType.local_low_config: (LinuxDirectoryType.local_config, None, ('local_low', 'config')),
    AppDirectoryType.local_low_log: (LinuxDirectoryType.local_log, None, ('local_low', 'log')),
    AppDirectoryType.local_low_temp: (LinuxDirectoryType.local_temp, None, ('local_low', 'temp')),
    AppDirectoryType.local_low_runtime: (LinuxDirectoryType.local_runtime, None, ('local_low', 'runtime')),

    AppDirectoryType.roaming_data: (LinuxDirectoryType.local_data, None, ('roaming', 'data')),
    AppDirectoryType.roaming_static_data: (LinuxDirectoryType.local_static_data, None, ('roaming', 'static_data')),
    AppDirectoryType.roaming_cache: (LinuxDirectoryType.local_cache, None, ('roaming', 'cache')),
    AppDirectoryType.roaming_config: (LinuxDirectoryType.local_config, None, ('roaming', 'config')),
    AppDirectoryType.roaming_log: (LinuxDirectoryType.local_log, None, ('roaming', 'log')),
    AppDirectoryType.roaming_temp: (LinuxDirectoryType.local_temp, None, ('roaming', 'temp')),
    AppDirectoryType.roaming_runtime: (LinuxDirectoryType.local_runtime, None, ('roaming', 'runtime')),

    AppDirectoryType.program_data: (LinuxDirectoryType.program_data, None, 'data'),
    AppDirectoryType.program_static_data: (LinuxDirectoryType.program_static_data, None, 'static_data'),
    AppDirectoryType.program_cache: (LinuxDirectoryType.program_cache, None, 'cache'),
    AppDirectoryType.program_config: (LinuxDirectoryType.program_config, None, 'config'),
    AppDirectoryType.program_log: (LinuxDirectoryType.program_log, None, 'log'),
    AppDirectoryType.program_temp: (LinuxDirectoryType.program_temp, None, 'temp'),
    AppDirectoryType.program_runtime: (LinuxDirectoryType.program_runtime, None, 'runtime'),

    AppDirectoryType.program_files: (LinuxDirectoryType.program_files, None, None),
    AppDirectoryType.program_files_common: (LinuxDirectoryType.program_files, None, None),

    AppDirectoryType.user_profile_data: (LinuxDirectoryType.user_profile_data, None, None),
    AppDirectoryType.user_profile_program_files: (LinuxDirectoryType.local_data, None, None),
    AppDirectoryType.user_profile_program_files_common: (LinuxDirectoryType.local_data, None, None),
}


class AppDirPath(AppDirPathBase):
    def __init__(self, max_cache_size: Optional[int] = None) -> None:
        super().__init__(max_cache_size)
    
    def dir_type_mapping(self, dir_type: AppDirectoryType) -> DirTypeMappingItem:
        return dir_type_mapping[dir_type]
    
    def base_dir_id_to_path(self, base_dir_id: BaseDirID) -> str:
        result_path = None

        if LinuxDirectoryType.local_data == base_dir_id:
            result_path = os.environ.get('XDG_STATE_HOME', os.path.join(os.path.expanduser('~'), '.local', 'state'))
        elif LinuxDirectoryType.local_static_data == base_dir_id:
            result_path = os.environ.get('XDG_DATA_HOME', os.path.join(os.path.expanduser('~'), '.local', 'share'))
        elif LinuxDirectoryType.local_cache == base_dir_id:
            result_path = os.environ.get('XDG_CACHE_HOME', os.path.join(os.path.expanduser('~'), '.cache'))
        elif LinuxDirectoryType.local_config == base_dir_id:
            result_path = os.environ.get('XDG_CONFIG_HOME', os.path.join(os.path.expanduser('~'), '.config'))
        elif LinuxDirectoryType.local_log == base_dir_id:
            result_path = os.environ.get('XDG_STATE_HOME', os.path.join(os.path.expanduser('~'), '.local', 'state'))
        elif LinuxDirectoryType.local_temp == base_dir_id:
            result_path = os.environ.get('XDG_RUNTIME_DIR', os.path.join(os.path.expanduser('~'), '.tmp'))
        elif LinuxDirectoryType.local_runtime == base_dir_id:
            result_path = os.environ.get('XDG_RUNTIME_DIR', os.path.join(os.path.expanduser('~'), '.tmp'))
        elif LinuxDirectoryType.program_data == base_dir_id:
            result_path = '/var/lib'
        elif LinuxDirectoryType.program_static_data == base_dir_id:
            result_path = '/usr/share'
        elif LinuxDirectoryType.program_cache == base_dir_id:
            result_path = '/var/cache'
        elif LinuxDirectoryType.program_config == base_dir_id:
            result_path = '/etc'
        elif LinuxDirectoryType.program_log == base_dir_id:
            result_path = '/var/log'
        elif LinuxDirectoryType.program_temp == base_dir_id:
            result_path = '/tmp'
        elif LinuxDirectoryType.program_runtime == base_dir_id:
            result_path = '/var/run'
        elif LinuxDirectoryType.program_files == base_dir_id:
            result_path = '/opt'
        elif LinuxDirectoryType.user_profile_data == base_dir_id:
            result_path = os.path.expanduser('~')
        elif LinuxDirectoryType.user_profile_program_files == base_dir_id:
            result_path = os.environ.get('XDG_DATA_HOME', os.path.join(os.path.expanduser('~'), '.local', 'share'))
        elif LinuxDirectoryType.user_profile_program_files_common == base_dir_id:
            result_path = os.environ.get('XDG_DATA_HOME', os.path.join(os.path.expanduser('~'), '.local', 'share'))

        return result_path



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'bytes__to__hex_string',
    'hex_string__to__bytes',
    'bytes__to__solid_hex_string',
    'solid_hex_string__to__bytes',
    'hex_dword_to_int',
    'int_to_hex_dword',
]


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.system import PLATFORM_NAME
from cengal.code_flow_control.none_or import none_or

import binascii


# HEX STRING - BYTES:


def bytes__to__hex_string(input_data: bytes, delimiter: str=None)->str:
    delimiter = none_or(delimiter, ' ')
    # 48508.80358249831 inputs per second
    fake_start = '00'
    fake_start_len = len(fake_start)
    hex_string = fake_start + binascii.hexlify(input_data).decode()
    result = delimiter.join(map(''.join, zip(*[iter(hex_string)]*2)))[fake_start_len:]
    result = result.strip()
    return result

# Very high memory consumption! Use .replace(' ', '') and then solid_hex_string__to__bytes() instead.
if 'PyPy' == PLATFORM_NAME:
    def hex_string__to__bytes(input_data: str, delimiter: str=None)->bytes:
        delimiter = none_or(delimiter, ' ')
        result = b''.join(binascii.unhexlify(b.encode()) for b in input_data.split(delimiter))
        return result
else:
    def hex_string__to__bytes(input_data: str, delimiter: str=None)->bytes:
        delimiter = none_or(delimiter, ' ')
        result = b''.join(binascii.unhexlify(b) for b in input_data.split(delimiter))
        return result


def bytes__to__solid_hex_string(input_data: bytes)->str:
    return binascii.hexlify(input_data).decode()

if 'PyPy' == PLATFORM_NAME:
    def solid_hex_string__to__bytes(input_data: str)->bytes:
        return binascii.unhexlify(input_data.encode())
else:
    def solid_hex_string__to__bytes(input_data: str)->bytes:
        return binascii.unhexlify(input_data)


# HEX DWORD:


def hex_dword_to_int(dword: str, byteorder=None, signed=False)->bytes:
    """
    "0x008A151D" is big endian; so dword = '008A151D', byteorder = 'big'
    :param dword: str(); '008A151D'
    :param byteorder: str(); 'big' / 'little'
    :param signed: bool();
    :return: 9049373
    """
    byteorder = byteorder or 'big'
    bin_dword = binascii.unhexlify(dword)
    result = int.from_bytes(bin_dword, byteorder=byteorder, signed=signed)
    return result


def int_to_hex_dword(int_value, byteorder=None, signed=False):
    byteorder = byteorder or 'big'
    result = (binascii.hexlify(int_value.to_bytes(4, byteorder=byteorder, signed=signed))).decode()
    return result



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


IF PYXD_UNAME_SYSNAME == "Windows":
    cdef extern from "Windows.h":
        void MemoryBarrier()

    
    cpdef void memory_barrier():
        MemoryBarrier()
    
    
    full_memory_barrier = memory_barrier

ELSE:
    cdef extern from "stdatomic.h":
        void atomic_thread_fence(int)


    cdef int _MEMORY_ORDER_RELAXED = 0
    cdef int _MEMORY_ORDER_CONSUME = 1
    cdef int _MEMORY_ORDER_ACQUIRE = 2
    cdef int _MEMORY_ORDER_RELEASE = 3
    cdef int _MEMORY_ORDER_ACQ_REL = 4
    cdef int _MEMORY_ORDER_SEQ_CST = 5


    MEMORY_ORDER_RELAXED = _MEMORY_ORDER_RELAXED
    MEMORY_ORDER_CONSUME = _MEMORY_ORDER_CONSUME
    MEMORY_ORDER_ACQUIRE = _MEMORY_ORDER_ACQUIRE
    MEMORY_ORDER_RELEASE = _MEMORY_ORDER_RELEASE
    MEMORY_ORDER_ACQ_REL = _MEMORY_ORDER_ACQ_REL
    MEMORY_ORDER_SEQ_CST = _MEMORY_ORDER_SEQ_CST


    cpdef void py_atomic_thread_fence(int order):
        cdef int c_order = <int>order
        atomic_thread_fence(c_order)


    cpdef void py_atomic_thread_fence__memory_order_relaxed():
        atomic_thread_fence(_MEMORY_ORDER_RELAXED)


    cpdef void py_atomic_thread_fence__memory_order_consume():
        atomic_thread_fence(_MEMORY_ORDER_CONSUME)


    cpdef void py_atomic_thread_fence__memory_order_acquire():
        atomic_thread_fence(_MEMORY_ORDER_ACQUIRE)


    cpdef void py_atomic_thread_fence__memory_order_release():
        atomic_thread_fence(_MEMORY_ORDER_RELEASE)


    cpdef void py_atomic_thread_fence__memory_order_acq_rel():
        atomic_thread_fence(_MEMORY_ORDER_ACQ_REL)


    cpdef void py_atomic_thread_fence__memory_order_seq_cst():
        atomic_thread_fence(_MEMORY_ORDER_SEQ_CST)


    full_memory_barrier = py_atomic_thread_fence__memory_order_seq_cst


IF (PYXD_UNAME_SYSNAME == "Windows") and (PYXD_UNAME_MACHINE in ("x86_64", "x86", "i386", "i686", "AMD64")):
    cdef extern from "<emmintrin.h>" nogil:
        void _mm_mfence()


    cdef extern from "<emmintrin.h>" nogil:
        void _mm_sfence()


    cdef extern from "<emmintrin.h>" nogil:
        void _mm_lfence()


    cdef extern from "<immintrin.h>" nogil:
        void _mm_pause()


    cdef extern from "emmintrin.h":
        void _mm_clflush(void* p)


    cpdef void mm_mfence():
        _mm_mfence()


    cpdef void mm_sfence():
        _mm_sfence()


    cpdef void mm_lfence():
        _mm_lfence()


    cpdef void mm_pause():
        _mm_pause()


    cpdef void mm_clflush(int p):
        _mm_clflush(<void*> p)

    
    cpdef void sync_synchronize():
        MemoryBarrier()


    full_memory_barrier = memory_barrier

ELIF (PYXD_UNAME_SYSNAME in ("Linux", "Darwin")) and (PYXD_UNAME_MACHINE in ("x86_64", "x86", "i386", "i686", "AMD64")):
    cdef extern from "<emmintrin.h>" nogil:
        void _mm_mfence()


    cdef extern from "<emmintrin.h>" nogil:
        void _mm_sfence()


    cdef extern from "<emmintrin.h>" nogil:
        void _mm_lfence()


    cdef extern from "<immintrin.h>" nogil:
        void _mm_pause()


    cdef extern from "<emmintrin.h>" nogil:
        void _mm_clflush(void* p)


    cdef extern from * nogil:
        void __sync_synchronize()


    # cdef extern from "stdlib.h" nogil:
    #     void __clear_cache(char* beg, char* end)


    cpdef void mm_mfence():
        _mm_mfence()


    cpdef void mm_sfence():
        _mm_sfence()


    cpdef void mm_lfence():
        _mm_lfence()


    cpdef void mm_pause():
        _mm_pause()

    
    cpdef void memory_barrier():
        _mm_mfence()


    cpdef void mm_clflush(int p):
        _mm_clflush(<void*> p)

    
    cpdef void sync_synchronize():
        __sync_synchronize()


    # # An alternative to mm_clflush() in GCC and Clang compilers
    # cpdef void clear_cache(int beg, int end):
    #     __clear_cache(<char*>beg, <char*>end)


    full_memory_barrier = mm_mfence

ELIF (PYXD_UNAME_SYSNAME == 'Windows') and PYXD_UNAME_MACHINE.startswith("arm"):
    cdef extern from "arm64_neon.h" nogil:
        void __iso_volatile_store16(void* p, unsigned long long val)


    # An alternative to mm_clflush() for ARM in Visual Studio compiler
    cpdef void iso_volatile_store16(int p, unsigned long long val):
        __iso_volatile_store16(<void*> p, val)


    full_memory_barrier = memory_barrier


    cpdef void mm_mfence():
        MemoryBarrier()


    cpdef void mm_sfence():
        MemoryBarrier()


    cpdef void mm_lfence():
        MemoryBarrier()


    cpdef void mm_pause():
        pass

    
    cpdef void sync_synchronize():
        MemoryBarrier()


    full_memory_barrier = memory_barrier

ELIF PYXD_UNAME_SYSNAME in ("Linux", "Darwin") and PYXD_UNAME_MACHINE.startswith("arm"):
    cdef extern from * nogil:
        void __sync_synchronize()

    
    cpdef void sync_synchronize():
        __sync_synchronize()


    # cdef extern from "stdlib.h" nogil:
    #     void __clear_cache(char* beg, char* end)


    # # An alternative to mm_clflush() for ARM in GCC and Clang compilers
    # cpdef void clear_cache(int beg, int end):
    #     __clear_cache(<char*>beg, <char*>end)


    full_memory_barrier = sync_synchronize


    cpdef void mm_mfence():
        __sync_synchronize()


    cpdef void mm_sfence():
        __sync_synchronize()


    cpdef void mm_lfence():
        __sync_synchronize()


    cpdef void mm_pause():
        pass

    
    cpdef void memory_barrier():
        __sync_synchronize()

ELIF PYXD_UNAME_SYSNAME == "Emscripten":
    cdef extern from * nogil:
        void __sync_synchronize()

    
    cpdef void sync_synchronize():
        __sync_synchronize()


    cdef extern from "emscripten/atomics.h" nogil:
        void emscripten_atomic_fence()


    # Probably can be used as an alternative to mm_clflush() for Emscripten
    cpdef void py_emscripten_atomic_fence():
        emscripten_atomic_fence()

    
    full_memory_barrier = emscripten_atomic_fence


    cpdef void mm_mfence():
        __sync_synchronize()


    cpdef void mm_sfence():
        __sync_synchronize()


    cpdef void mm_lfence():
        __sync_synchronize()


    cpdef void mm_pause():
        pass

    
    cpdef void memory_barrier():
        __sync_synchronize()

ELSE:
    full_memory_barrier = py_atomic_thread_fence__memory_order_seq_cst


    cpdef void mm_mfence():
        atomic_thread_fence(_MEMORY_ORDER_SEQ_CST)


    cpdef void mm_sfence():
        atomic_thread_fence(_MEMORY_ORDER_SEQ_CST)


    cpdef void mm_lfence():
        atomic_thread_fence(_MEMORY_ORDER_SEQ_CST)


    cpdef void mm_pause():
        pass

    
    cpdef void sync_synchronize():
        atomic_thread_fence(_MEMORY_ORDER_SEQ_CST)

    
    cpdef void memory_barrier():
        atomic_thread_fence(_MEMORY_ORDER_SEQ_CST)


IF PYXD_UNAME_SYSNAME == "Windows":
    full_memory_barrier = memory_barrier
